﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace YOYO_GUI1 {
    struct Pair {
        public byte[,] p1;
        public byte[,] p2;
        public byte[,] c1;
        public byte[,] c2;
    }
    class Rijndael {
        Random rand;

        public Rijndael() {
            rand = new Random();
        }

        // Math ------------------------------------------------------------------------------------------------------------------
        public double Log2(double n) { return Math.Log(n) / Math.Log(2.0); }  // log(n)/log(2) is log2.         
        public byte Mul(int a, int b) { //multiply two elements of GF(2^m)
            if (a != 0 && b != 0) return Boxes.Alogtable[(Boxes.Logtable[a] + Boxes.Logtable[b]) % 255];
            else return 0;
        }
        public int Mod(int a, int b) {
            int r = a % b;
            return r < 0 ? r + b : r;
        }
        public int Inv(int a) { //modifisert til byte
            if (a == 0) return 0;       /* as needed in the context of AES */
            int x = Boxes.Logtable[a];   /* x       is in range [0..255] */
            int log_inv = 255 - x;  /* log_inv is in range [0..255] */
            return Boxes.Alogtable[log_inv];
        }

        // Key functions ---------------------------------------------------------------------------------------------------------        
        public List<byte[,]> KeySchedule(byte[,] k) {
            List<byte[,]> roundKeyList = new List<byte[,]>();
            roundKeyList.Add(k);
            for (int keyNum = 0; keyNum < 10; keyNum++) {
                byte[] rW1 = new byte[4] { k[0, 0], k[1, 0], k[2, 0], k[3, 0] };
                byte[] rW2 = new byte[4] { k[0, 1], k[1, 1], k[2, 1], k[3, 1] };
                byte[] rW3 = new byte[4] { k[0, 2], k[1, 2], k[2, 2], k[3, 2] };
                byte[] rW4 = new byte[4] { k[0, 3], k[1, 3], k[2, 3], k[3, 3] };

                byte[] rW4Flip = new byte[4] { k[1, 3], k[2, 3], k[3, 3], k[0, 3] }; //flipped rW4                                
                byte[] sub_rW4Flip = new byte[4] { Boxes.s_box[rW4Flip[0]], Boxes.s_box[rW4Flip[1]], Boxes.s_box[rW4Flip[2]], Boxes.s_box[rW4Flip[3]] };

                byte[] kW1 = new byte[4]; for (int i = 0; i < 4; i++) { kW1[i] = (byte)(rW1[i] ^ sub_rW4Flip[i] ^ Boxes.Rcon[i, keyNum]); }
                byte[] kW2 = new byte[4]; for (int i = 0; i < 4; i++) { kW2[i] = (byte)(rW2[i] ^ kW1[i]); }
                byte[] kW3 = new byte[4]; for (int i = 0; i < 4; i++) { kW3[i] = (byte)(rW3[i] ^ kW2[i]); }
                byte[] kW4 = new byte[4]; for (int i = 0; i < 4; i++) { kW4[i] = (byte)(rW4[i] ^ kW3[i]); }

                byte[,] roundKey = new byte[4, 4] {
                    { kW1[0], kW2[0], kW3[0], kW4[0] },
                    { kW1[1], kW2[1], kW3[1], kW4[1] },
                    { kW1[2], kW2[2], kW3[2], kW4[2] },
                    { kW1[3], kW2[3], kW3[3], kW4[3] },
                };
                roundKeyList.Add(roundKey);
                k = roundKey;
            }
            return roundKeyList;
        }
        public byte[,] KeySched1(byte[,] k) {
            byte[] rW1 = new byte[4] { k[0, 0], k[1, 0], k[2, 0], k[3, 0] };
            byte[] rW2 = new byte[4] { k[0, 1], k[1, 1], k[2, 1], k[3, 1] };
            byte[] rW3 = new byte[4] { k[0, 2], k[1, 2], k[2, 2], k[3, 2] };
            byte[] rW4 = new byte[4] { k[0, 3], k[1, 3], k[2, 3], k[3, 3] };

            byte[] rW4Flip = new byte[4] { k[1, 3], k[2, 3], k[3, 3], k[0, 3] }; //flipped rW4                                
            byte[] sub_rW4Flip = new byte[4] { Boxes.s_box[rW4Flip[0]], Boxes.s_box[rW4Flip[1]], Boxes.s_box[rW4Flip[2]], Boxes.s_box[rW4Flip[3]] };

            byte[] kW1 = new byte[4]; for (int i = 0; i < 4; i++) { kW1[i] = (byte)(rW1[i] ^ sub_rW4Flip[i] ^ Boxes.Rcon[i, 0]); }
            byte[] kW2 = new byte[4]; for (int i = 0; i < 4; i++) { kW2[i] = (byte)(rW2[i] ^ kW1[i]); }
            byte[] kW3 = new byte[4]; for (int i = 0; i < 4; i++) { kW3[i] = (byte)(rW3[i] ^ kW2[i]); }
            byte[] kW4 = new byte[4]; for (int i = 0; i < 4; i++) { kW4[i] = (byte)(rW4[i] ^ kW3[i]); }

            byte[,] roundKey = new byte[4, 4] {
                    { kW1[0], kW2[0], kW3[0], kW4[0] },
                    { kW1[1], kW2[1], kW3[1], kW4[1] },
                    { kW1[2], kW2[2], kW3[2], kW4[2] },
                    { kW1[3], kW2[3], kW3[3], kW4[3] },
                };
            return roundKey;
        }
        public List<byte[,]> KeySched5(byte[,] k) {
            List<byte[,]> roundKeyList = new List<byte[,]>();

            roundKeyList.Add(k);

            for (int keyNum = 0; keyNum < 5; keyNum++) {
                byte[] rW1 = new byte[4] { k[0, 0], k[1, 0], k[2, 0], k[3, 0] };
                byte[] rW2 = new byte[4] { k[0, 1], k[1, 1], k[2, 1], k[3, 1] };
                byte[] rW3 = new byte[4] { k[0, 2], k[1, 2], k[2, 2], k[3, 2] };
                byte[] rW4 = new byte[4] { k[0, 3], k[1, 3], k[2, 3], k[3, 3] };

                byte[] rW4Flip = new byte[4] { k[1, 3], k[2, 3], k[3, 3], k[0, 3] }; //flipped rW4                                
                byte[] sub_rW4Flip = new byte[4] { Boxes.s_box[rW4Flip[0]], Boxes.s_box[rW4Flip[1]], Boxes.s_box[rW4Flip[2]], Boxes.s_box[rW4Flip[3]] };

                byte[] kW1 = new byte[4]; for (int i = 0; i < 4; i++) { kW1[i] = (byte)(rW1[i] ^ sub_rW4Flip[i] ^ Boxes.Rcon[i, keyNum]); }
                byte[] kW2 = new byte[4]; for (int i = 0; i < 4; i++) { kW2[i] = (byte)(rW2[i] ^ kW1[i]); }
                byte[] kW3 = new byte[4]; for (int i = 0; i < 4; i++) { kW3[i] = (byte)(rW3[i] ^ kW2[i]); }
                byte[] kW4 = new byte[4]; for (int i = 0; i < 4; i++) { kW4[i] = (byte)(rW4[i] ^ kW3[i]); }

                byte[,] roundKey = new byte[4, 4] {
                    { kW1[0], kW2[0], kW3[0], kW4[0] },
                    { kW1[1], kW2[1], kW3[1], kW4[1] },
                    { kW1[2], kW2[2], kW3[2], kW4[2] },
                    { kW1[3], kW2[3], kW3[3], kW4[3] },
                };
                roundKeyList.Add(roundKey);
                k = roundKey;
            }
            return roundKeyList;
        }
        public byte[,] GetKeyFrom_rk1(byte[,] rk) {

            byte[] rk0 = new byte[4] { rk[0, 0], rk[1, 0], rk[2, 0], rk[3, 0] };
            byte[] rk1 = new byte[4] { rk[0, 1], rk[1, 1], rk[2, 1], rk[3, 1] };
            byte[] rk2 = new byte[4] { rk[0, 2], rk[1, 2], rk[2, 2], rk[3, 2] };
            byte[] rk3 = new byte[4] { rk[0, 3], rk[1, 3], rk[2, 3], rk[3, 3] };


            byte[] k3 = new byte[4]; for (int i = 0; i < 4; i++) { k3[i] = (byte)(rk2[i] ^ rk3[i]); }
            byte[] k2 = new byte[4]; for (int i = 0; i < 4; i++) { k2[i] = (byte)(rk1[i] ^ rk2[i]); }
            byte[] k1 = new byte[4]; for (int i = 0; i < 4; i++) { k1[i] = (byte)(rk0[i] ^ rk1[i]); }

            byte[] k3Flip = new byte[4] { k3[1], k3[2], k3[3], k3[0] };
            byte[] sub_k3Flip = new byte[4] { Boxes.s_box[k3Flip[0]], Boxes.s_box[k3Flip[1]], Boxes.s_box[k3Flip[2]], Boxes.s_box[k3Flip[3]] };
            byte[] k0 = new byte[4]; for (int i = 0; i < 4; i++) { k0[i] = (byte)(rk0[i] ^ sub_k3Flip[i] ^ Boxes.Rcon[i, 0]); }

            byte[,] k = new byte[4, 4] {
                    { k0[0], k1[0], k2[0], k3[0] },
                    { k0[1], k1[1], k2[1], k3[1] },
                    { k0[2], k1[2], k2[2], k3[2] },
                    { k0[3], k1[3], k2[3], k3[3] },
            };
            return k;
        }
        public byte[,] GetKeyFrom_rk2(byte[,] rk) {
            byte[,] k = new byte[4, 4];
            for (int round = 1; round >= 0; round--) {
                byte[] rk0 = new byte[4] { rk[0, 0], rk[1, 0], rk[2, 0], rk[3, 0] };
                byte[] rk1 = new byte[4] { rk[0, 1], rk[1, 1], rk[2, 1], rk[3, 1] };
                byte[] rk2 = new byte[4] { rk[0, 2], rk[1, 2], rk[2, 2], rk[3, 2] };
                byte[] rk3 = new byte[4] { rk[0, 3], rk[1, 3], rk[2, 3], rk[3, 3] };


                byte[] k3 = new byte[4]; for (int i = 0; i < 4; i++) { k3[i] = (byte)(rk2[i] ^ rk3[i]); }
                byte[] k2 = new byte[4]; for (int i = 0; i < 4; i++) { k2[i] = (byte)(rk1[i] ^ rk2[i]); }
                byte[] k1 = new byte[4]; for (int i = 0; i < 4; i++) { k1[i] = (byte)(rk0[i] ^ rk1[i]); }

                byte[] k3Flip = new byte[4] { k3[1], k3[2], k3[3], k3[0] };
                byte[] sub_k3Flip = new byte[4] { Boxes.s_box[k3Flip[0]], Boxes.s_box[k3Flip[1]], Boxes.s_box[k3Flip[2]], Boxes.s_box[k3Flip[3]] };
                byte[] k0 = new byte[4]; for (int i = 0; i < 4; i++) { k0[i] = (byte)(rk0[i] ^ sub_k3Flip[i] ^ Boxes.Rcon[i, round]); }

                k = new byte[4, 4] {
                    { k0[0], k1[0], k2[0], k3[0] },
                    { k0[1], k1[1], k2[1], k3[1] },
                    { k0[2], k1[2], k2[2], k3[2] },
                    { k0[3], k1[3], k2[3], k3[3] },
                };
                rk = copy(k);
            }
            return k;
        }
        public byte[,] GetKeyFrom_rk3(byte[,] rk) {
            byte[,] k = new byte[4, 4];
            for (int round = 2; round >= 0; round--) {
                byte[] rk0 = new byte[4] { rk[0, 0], rk[1, 0], rk[2, 0], rk[3, 0] };
                byte[] rk1 = new byte[4] { rk[0, 1], rk[1, 1], rk[2, 1], rk[3, 1] };
                byte[] rk2 = new byte[4] { rk[0, 2], rk[1, 2], rk[2, 2], rk[3, 2] };
                byte[] rk3 = new byte[4] { rk[0, 3], rk[1, 3], rk[2, 3], rk[3, 3] };


                byte[] k3 = new byte[4]; for (int i = 0; i < 4; i++) { k3[i] = (byte)(rk2[i] ^ rk3[i]); }
                byte[] k2 = new byte[4]; for (int i = 0; i < 4; i++) { k2[i] = (byte)(rk1[i] ^ rk2[i]); }
                byte[] k1 = new byte[4]; for (int i = 0; i < 4; i++) { k1[i] = (byte)(rk0[i] ^ rk1[i]); }

                byte[] k3Flip = new byte[4] { k3[1], k3[2], k3[3], k3[0] };
                byte[] sub_k3Flip = new byte[4] { Boxes.s_box[k3Flip[0]], Boxes.s_box[k3Flip[1]], Boxes.s_box[k3Flip[2]], Boxes.s_box[k3Flip[3]] };
                byte[] k0 = new byte[4]; for (int i = 0; i < 4; i++) { k0[i] = (byte)(rk0[i] ^ sub_k3Flip[i] ^ Boxes.Rcon[i, round]); }

                k = new byte[4, 4] {
                    { k0[0], k1[0], k2[0], k3[0] },
                    { k0[1], k1[1], k2[1], k3[1] },
                    { k0[2], k1[2], k2[2], k3[2] },
                    { k0[3], k1[3], k2[3], k3[3] },
                };
                rk = copy(k);
            }
            return k;
        }
        public byte[,] GetKeyFrom_rk4(byte[,] rk) {
            byte[,] k = new byte[4, 4];
            for (int round = 3; round >= 0; round--) {
                byte[] rk0 = new byte[4] { rk[0, 0], rk[1, 0], rk[2, 0], rk[3, 0] };
                byte[] rk1 = new byte[4] { rk[0, 1], rk[1, 1], rk[2, 1], rk[3, 1] };
                byte[] rk2 = new byte[4] { rk[0, 2], rk[1, 2], rk[2, 2], rk[3, 2] };
                byte[] rk3 = new byte[4] { rk[0, 3], rk[1, 3], rk[2, 3], rk[3, 3] };


                byte[] k3 = new byte[4]; for (int i = 0; i < 4; i++) { k3[i] = (byte)(rk2[i] ^ rk3[i]); }
                byte[] k2 = new byte[4]; for (int i = 0; i < 4; i++) { k2[i] = (byte)(rk1[i] ^ rk2[i]); }
                byte[] k1 = new byte[4]; for (int i = 0; i < 4; i++) { k1[i] = (byte)(rk0[i] ^ rk1[i]); }

                byte[] k3Flip = new byte[4] { k3[1], k3[2], k3[3], k3[0] };
                byte[] sub_k3Flip = new byte[4] { Boxes.s_box[k3Flip[0]], Boxes.s_box[k3Flip[1]], Boxes.s_box[k3Flip[2]], Boxes.s_box[k3Flip[3]] };
                byte[] k0 = new byte[4]; for (int i = 0; i < 4; i++) { k0[i] = (byte)(rk0[i] ^ sub_k3Flip[i] ^ Boxes.Rcon[i, round]); }

                k = new byte[4, 4] {
                    { k0[0], k1[0], k2[0], k3[0] },
                    { k0[1], k1[1], k2[1], k3[1] },
                    { k0[2], k1[2], k2[2], k3[2] },
                    { k0[3], k1[3], k2[3], k3[3] },
                };
                rk = copy(k);
            }
            return k;
        }
        public byte[,] GetKeyFrom_rk5(byte[,] rk) {
            byte[,] k = new byte[4, 4];
            for (int round = 4; round >= 0; round--) {
                byte[] rk0 = new byte[4] { rk[0, 0], rk[1, 0], rk[2, 0], rk[3, 0] };
                byte[] rk1 = new byte[4] { rk[0, 1], rk[1, 1], rk[2, 1], rk[3, 1] };
                byte[] rk2 = new byte[4] { rk[0, 2], rk[1, 2], rk[2, 2], rk[3, 2] };
                byte[] rk3 = new byte[4] { rk[0, 3], rk[1, 3], rk[2, 3], rk[3, 3] };


                byte[] k3 = new byte[4]; for (int i = 0; i < 4; i++) { k3[i] = (byte)(rk2[i] ^ rk3[i]); }
                byte[] k2 = new byte[4]; for (int i = 0; i < 4; i++) { k2[i] = (byte)(rk1[i] ^ rk2[i]); }
                byte[] k1 = new byte[4]; for (int i = 0; i < 4; i++) { k1[i] = (byte)(rk0[i] ^ rk1[i]); }

                byte[] k3Flip = new byte[4] { k3[1], k3[2], k3[3], k3[0] };
                byte[] sub_k3Flip = new byte[4] { Boxes.s_box[k3Flip[0]], Boxes.s_box[k3Flip[1]], Boxes.s_box[k3Flip[2]], Boxes.s_box[k3Flip[3]] };
                byte[] k0 = new byte[4]; for (int i = 0; i < 4; i++) { k0[i] = (byte)(rk0[i] ^ sub_k3Flip[i] ^ Boxes.Rcon[i, round]); }

                k = new byte[4, 4] {
                    { k0[0], k1[0], k2[0], k3[0] },
                    { k0[1], k1[1], k2[1], k3[1] },
                    { k0[2], k1[2], k2[2], k3[2] },
                    { k0[3], k1[3], k2[3], k3[3] },
                };
                rk = copy(k);
            }
            return k;
        }
        public byte[,] GetKeyFrom_rk6(byte[,] rk) {
            byte[,] k = new byte[4, 4];
            for (int round = 5; round >= 0; round--) {
                byte[] rk0 = new byte[4] { rk[0, 0], rk[1, 0], rk[2, 0], rk[3, 0] };
                byte[] rk1 = new byte[4] { rk[0, 1], rk[1, 1], rk[2, 1], rk[3, 1] };
                byte[] rk2 = new byte[4] { rk[0, 2], rk[1, 2], rk[2, 2], rk[3, 2] };
                byte[] rk3 = new byte[4] { rk[0, 3], rk[1, 3], rk[2, 3], rk[3, 3] };


                byte[] k3 = new byte[4]; for (int i = 0; i < 4; i++) { k3[i] = (byte)(rk2[i] ^ rk3[i]); }
                byte[] k2 = new byte[4]; for (int i = 0; i < 4; i++) { k2[i] = (byte)(rk1[i] ^ rk2[i]); }
                byte[] k1 = new byte[4]; for (int i = 0; i < 4; i++) { k1[i] = (byte)(rk0[i] ^ rk1[i]); }

                byte[] k3Flip = new byte[4] { k3[1], k3[2], k3[3], k3[0] };
                byte[] sub_k3Flip = new byte[4] { Boxes.s_box[k3Flip[0]], Boxes.s_box[k3Flip[1]], Boxes.s_box[k3Flip[2]], Boxes.s_box[k3Flip[3]] };
                byte[] k0 = new byte[4]; for (int i = 0; i < 4; i++) { k0[i] = (byte)(rk0[i] ^ sub_k3Flip[i] ^ Boxes.Rcon[i, round]); }

                k = new byte[4, 4] {
                    { k0[0], k1[0], k2[0], k3[0] },
                    { k0[1], k1[1], k2[1], k3[1] },
                    { k0[2], k1[2], k2[2], k3[2] },
                    { k0[3], k1[3], k2[3], k3[3] },
                };
                rk = copy(k);
            }
            return k;
        }
        public byte[,] GetKeyFrom_rk7(byte[,] rk) {
            byte[,] k = new byte[4, 4];
            for (int round = 6; round >= 0; round--) {
                byte[] rk0 = new byte[4] { rk[0, 0], rk[1, 0], rk[2, 0], rk[3, 0] };
                byte[] rk1 = new byte[4] { rk[0, 1], rk[1, 1], rk[2, 1], rk[3, 1] };
                byte[] rk2 = new byte[4] { rk[0, 2], rk[1, 2], rk[2, 2], rk[3, 2] };
                byte[] rk3 = new byte[4] { rk[0, 3], rk[1, 3], rk[2, 3], rk[3, 3] };


                byte[] k3 = new byte[4]; for (int i = 0; i < 4; i++) { k3[i] = (byte)(rk2[i] ^ rk3[i]); }
                byte[] k2 = new byte[4]; for (int i = 0; i < 4; i++) { k2[i] = (byte)(rk1[i] ^ rk2[i]); }
                byte[] k1 = new byte[4]; for (int i = 0; i < 4; i++) { k1[i] = (byte)(rk0[i] ^ rk1[i]); }

                byte[] k3Flip = new byte[4] { k3[1], k3[2], k3[3], k3[0] };
                byte[] sub_k3Flip = new byte[4] { Boxes.s_box[k3Flip[0]], Boxes.s_box[k3Flip[1]], Boxes.s_box[k3Flip[2]], Boxes.s_box[k3Flip[3]] };
                byte[] k0 = new byte[4]; for (int i = 0; i < 4; i++) { k0[i] = (byte)(rk0[i] ^ sub_k3Flip[i] ^ Boxes.Rcon[i, round]); }

                k = new byte[4, 4] {
                    { k0[0], k1[0], k2[0], k3[0] },
                    { k0[1], k1[1], k2[1], k3[1] },
                    { k0[2], k1[2], k2[2], k3[2] },
                    { k0[3], k1[3], k2[3], k3[3] },
                };
                rk = copy(k);
            }
            return k;
        }
        public byte[,] GetKeyFrom_rk8(byte[,] rk) {
            byte[,] k = new byte[4, 4];
            for (int round = 7; round >= 0; round--) {
                byte[] rk0 = new byte[4] { rk[0, 0], rk[1, 0], rk[2, 0], rk[3, 0] };
                byte[] rk1 = new byte[4] { rk[0, 1], rk[1, 1], rk[2, 1], rk[3, 1] };
                byte[] rk2 = new byte[4] { rk[0, 2], rk[1, 2], rk[2, 2], rk[3, 2] };
                byte[] rk3 = new byte[4] { rk[0, 3], rk[1, 3], rk[2, 3], rk[3, 3] };


                byte[] k3 = new byte[4]; for (int i = 0; i < 4; i++) { k3[i] = (byte)(rk2[i] ^ rk3[i]); }
                byte[] k2 = new byte[4]; for (int i = 0; i < 4; i++) { k2[i] = (byte)(rk1[i] ^ rk2[i]); }
                byte[] k1 = new byte[4]; for (int i = 0; i < 4; i++) { k1[i] = (byte)(rk0[i] ^ rk1[i]); }

                byte[] k3Flip = new byte[4] { k3[1], k3[2], k3[3], k3[0] };
                byte[] sub_k3Flip = new byte[4] { Boxes.s_box[k3Flip[0]], Boxes.s_box[k3Flip[1]], Boxes.s_box[k3Flip[2]], Boxes.s_box[k3Flip[3]] };
                byte[] k0 = new byte[4]; for (int i = 0; i < 4; i++) { k0[i] = (byte)(rk0[i] ^ sub_k3Flip[i] ^ Boxes.Rcon[i, round]); }

                k = new byte[4, 4] {
                    { k0[0], k1[0], k2[0], k3[0] },
                    { k0[1], k1[1], k2[1], k3[1] },
                    { k0[2], k1[2], k2[2], k3[2] },
                    { k0[3], k1[3], k2[3], k3[3] },
                };
                rk = copy(k);
            }
            return k;
        }
        public byte[,] GetKeyFrom_rk9(byte[,] rk) {
            byte[,] k = new byte[4, 4];
            for (int round = 8; round >= 0; round--) {
                byte[] rk0 = new byte[4] { rk[0, 0], rk[1, 0], rk[2, 0], rk[3, 0] };
                byte[] rk1 = new byte[4] { rk[0, 1], rk[1, 1], rk[2, 1], rk[3, 1] };
                byte[] rk2 = new byte[4] { rk[0, 2], rk[1, 2], rk[2, 2], rk[3, 2] };
                byte[] rk3 = new byte[4] { rk[0, 3], rk[1, 3], rk[2, 3], rk[3, 3] };


                byte[] k3 = new byte[4]; for (int i = 0; i < 4; i++) { k3[i] = (byte)(rk2[i] ^ rk3[i]); }
                byte[] k2 = new byte[4]; for (int i = 0; i < 4; i++) { k2[i] = (byte)(rk1[i] ^ rk2[i]); }
                byte[] k1 = new byte[4]; for (int i = 0; i < 4; i++) { k1[i] = (byte)(rk0[i] ^ rk1[i]); }

                byte[] k3Flip = new byte[4] { k3[1], k3[2], k3[3], k3[0] };
                byte[] sub_k3Flip = new byte[4] { Boxes.s_box[k3Flip[0]], Boxes.s_box[k3Flip[1]], Boxes.s_box[k3Flip[2]], Boxes.s_box[k3Flip[3]] };
                byte[] k0 = new byte[4]; for (int i = 0; i < 4; i++) { k0[i] = (byte)(rk0[i] ^ sub_k3Flip[i] ^ Boxes.Rcon[i, round]); }

                k = new byte[4, 4] {
                    { k0[0], k1[0], k2[0], k3[0] },
                    { k0[1], k1[1], k2[1], k3[1] },
                    { k0[2], k1[2], k2[2], k3[2] },
                    { k0[3], k1[3], k2[3], k3[3] },
                };
                rk = copy(k);
            }
            return k;
        }
        public byte[,] GetKeyFrom_rk10(byte[,] rk) {
            byte[,] k = new byte[4, 4];
            for (int round = 9; round >= 0; round--) {
                byte[] rk0 = new byte[4] { rk[0, 0], rk[1, 0], rk[2, 0], rk[3, 0] };
                byte[] rk1 = new byte[4] { rk[0, 1], rk[1, 1], rk[2, 1], rk[3, 1] };
                byte[] rk2 = new byte[4] { rk[0, 2], rk[1, 2], rk[2, 2], rk[3, 2] };
                byte[] rk3 = new byte[4] { rk[0, 3], rk[1, 3], rk[2, 3], rk[3, 3] };


                byte[] k3 = new byte[4]; for (int i = 0; i < 4; i++) { k3[i] = (byte)(rk2[i] ^ rk3[i]); }
                byte[] k2 = new byte[4]; for (int i = 0; i < 4; i++) { k2[i] = (byte)(rk1[i] ^ rk2[i]); }
                byte[] k1 = new byte[4]; for (int i = 0; i < 4; i++) { k1[i] = (byte)(rk0[i] ^ rk1[i]); }

                byte[] k3Flip = new byte[4] { k3[1], k3[2], k3[3], k3[0] };
                byte[] sub_k3Flip = new byte[4] { Boxes.s_box[k3Flip[0]], Boxes.s_box[k3Flip[1]], Boxes.s_box[k3Flip[2]], Boxes.s_box[k3Flip[3]] };
                byte[] k0 = new byte[4]; for (int i = 0; i < 4; i++) { k0[i] = (byte)(rk0[i] ^ sub_k3Flip[i] ^ Boxes.Rcon[i, round]); }

                k = new byte[4, 4] {
                    { k0[0], k1[0], k2[0], k3[0] },
                    { k0[1], k1[1], k2[1], k3[1] },
                    { k0[2], k1[2], k2[2], k3[2] },
                    { k0[3], k1[3], k2[3], k3[3] },
                };
                rk = copy(k);
            }
            return k;
        }

        public byte[,] AddRoundKey(byte[,] a, byte[,] rk) {
            byte[,] b = new byte[4, 4];
            for (int i = 0; i < 4; i++)
                for (int j = 0; j < 4; j++) b[i, j] = (byte)(a[i, j] ^ rk[i, j]);
            return b;
        }
        public byte[,] SubByte(byte[,] a) {
            byte[,] sub = new byte[4, 4];
            for (int i = 0; i < 4; i++)
                for (int j = 0; j < 4; j++)
                    sub[i, j] = Boxes.s_box[a[i, j]];
            return sub;
        }
        public byte[,] InvSubByte(byte[,] a) {
            byte[,] inv = new byte[4, 4];
            for (int i = 0; i < 4; i++)
                for (int j = 0; j < 4; j++)
                    inv[i, j] = Boxes.si_box[a[i, j]];
            return inv;
        }
        public byte[,] ShiftRows(byte[,] a) {
            byte[] tmp = new byte[4];
            byte[,] inv = copy(a);
            for (int i = 1; i < 4; i++) {
                for (int j = 0; j < 4; j++)
                    tmp[j] = a[i, (j + Boxes.shifts[0, i, 0]) % 4];
                for (int j = 0; j < 4; j++) inv[i, j] = tmp[j];
            }
            return inv;
        }
        public byte[,] InvShiftRows(byte[,] a) {
            byte[] tmp = new byte[4];
            byte[,] inv = copy(a);
            for (int i = 1; i < 4; i++) {
                for (int j = 0; j < 4; j++)
                    tmp[j] = a[i, (j + Boxes.shifts[0, i, 1]) % 4];
                for (int j = 0; j < 4; j++) inv[i, j] = tmp[j];
            }
            return inv;
        }
        public byte[,] MixColumns(byte[,] a) {
            byte[,] b = new byte[4, 4];
            for (int j = 0; j < 4; j++)
                for (int i = 0; i < 4; i++)
                    b[i, j] = (byte)(Mul(2, a[i, j]) ^ Mul(3, a[(i + 1) % 4, j]) ^ a[(i + 2) % 4, j] ^ a[(i + 3) % 4, j]);
            return b;
        }
        public byte[,] InvMixColumns(byte[,] a) {
            byte[,] b = new byte[4, 4];
            for (int j = 0; j < 4; j++)
                for (int i = 0; i < 4; i++)
                    b[i, j] = (byte)(Mul(0xe, a[i, j]) ^ Mul(0xb, a[(i + 1) % 4, j]) ^ Mul(0xd, a[(i + 2) % 4, j]) ^ Mul(0x9, a[(i + 3) % 4, j]));
            return b;
        }
        //---------------------------------------------------------------------
        public byte[,] ARK(byte[,] a, byte[,] rk) {
            byte[,] b = new byte[4, 4];
            for (int i = 0; i < 4; i++)
                for (int j = 0; j < 4; j++) b[i, j] = (byte)(a[i, j] ^ rk[i, j]);
            return b;
        }
        public byte[,] SB(byte[,] a) {
            byte[,] sub = new byte[4, 4];
            for (int i = 0; i < 4; i++)
                for (int j = 0; j < 4; j++)
                    sub[i, j] = Boxes.s_box[a[i, j]];
            return sub;
        }
        public byte[,] SR(byte[,] a) {
            byte[] tmp = new byte[4];
            byte[,] inv = copy(a);
            for (int i = 1; i < 4; i++) {
                for (int j = 0; j < 4; j++)
                    tmp[j] = a[i, (j + Boxes.shifts[0, i, 0]) % 4];
                for (int j = 0; j < 4; j++) inv[i, j] = tmp[j];
            }
            return inv;
        }
        public byte[,] MC(byte[,] a) {
            byte[,] b = new byte[4, 4];
            for (int j = 0; j < 4; j++)
                for (int i = 0; i < 4; i++)
                    b[i, j] = (byte)(Mul(2, a[i, j]) ^ Mul(3, a[(i + 1) % 4, j]) ^ a[(i + 2) % 4, j] ^ a[(i + 3) % 4, j]);
            return b;
        }
        public byte[,] SBi(byte[,] a) {
            byte[,] inv = new byte[4, 4];
            for (int i = 0; i < 4; i++)
                for (int j = 0; j < 4; j++)
                    inv[i, j] = Boxes.si_box[a[i, j]];
            return inv;
        }
        public byte[,] SRi(byte[,] a) {
            byte[] tmp = new byte[4];
            byte[,] inv = copy(a);
            for (int i = 1; i < 4; i++) {
                for (int j = 0; j < 4; j++)
                    tmp[j] = a[i, (j + Boxes.shifts[0, i, 1]) % 4];
                for (int j = 0; j < 4; j++) inv[i, j] = tmp[j];
            }
            return inv;
        }
        public byte[,] MCi(byte[,] a) {
            byte[,] b = new byte[4, 4];
            for (int j = 0; j < 4; j++)
                for (int i = 0; i < 4; i++)
                    b[i, j] = (byte)(Mul(0xe, a[i, j]) ^ Mul(0xb, a[(i + 1) % 4, j]) ^ Mul(0xd, a[(i + 2) % 4, j]) ^ Mul(0x9, a[(i + 3) % 4, j]));
            return b;
        }

        public byte[,] SBi_SRi_MCi(byte[,] a) {
            byte[,] b = MCi(a);
            byte[,] c = new byte[4, 4];
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) { c[i, j] = Boxes.si_box[b[i, (j + ((4 - i) % 4)) % 4]]; }
            }
            /*
            for (int j = 0; j < 4; j++) { c[0, j] = Boxes.si_box[b[0, j]]; }
            for (int j = 0; j < 4; j++) { c[1, j] = Boxes.si_box[b[1, (j + 3) % 4]]; }
            for (int j = 0; j < 4; j++) { c[2, j] = Boxes.si_box[b[2, (j + 2) % 4]]; }
            for (int j = 0; j < 4; j++) { c[3, j] = Boxes.si_box[b[3, (j + 1) % 4]]; }*/
            return c;
        }
        public byte[,] SBSRMC(byte[,] a) {
            byte[,] b = SB(a);
            b = SR(b);
            b = MC(b);
            return b;
        }
        public byte[,] MCiSRiSBi(byte[,] a) {
            byte[,] b = MCi(a);
            b = SRi(b);
            b = SBi(b);
            return b;
        }
        public byte[,] MCMCSRi(byte[,] a) {
            byte[,] b = copy(a);
            b = MC(b);
            b = MC(b);
            b = SRi(b);
            return b;
        }
        public byte[,] SRMCiMCi(byte[,] a) {
            byte[,] b = copy(a);
            b = SR(b);
            b = MCi(b);
            b = MCi(b);
            return b;
        }

        // Encryption -----------------------------------------------------------------------------------------------------------
        public byte[,] Encrypt(byte[,] a, List<byte[,]> rk) {
            a = AddRoundKey(a, rk[0]);
            for (int i = 0; i < 9; i++) {
                a = SubByte(a);
                a = ShiftRows(a);
                a = MixColumns(a);
                a = AddRoundKey(a, rk[i + 1]);
            }
            a = SubByte(a);
            a = ShiftRows(a);
            a = AddRoundKey(a, rk[10]);
            return a;
        }
        public byte[,] EncryptPartial(byte[,] a, List<byte[,]> rk, int rounds) {
            a = AddRoundKey(a, rk[0]);
            for (int r = 0; r < rounds; r++) {
                a = SubByte(a);
                a = ShiftRows(a);
                a = MixColumns(a);
                a = AddRoundKey(a, rk[r + 1]);
            }
            return a;
        }
        public byte[,] Enc(byte[,] a, List<byte[,]> rk, int rounds) {
            byte[,] b = copy(a);
            b = AddRoundKey(b, rk[0]);
            for (int r = 0; r < rounds; r++) {
                b = SubByte(b);
                b = ShiftRows(b);
                b = MixColumns(b);
                b = AddRoundKey(b, rk[r + 1]);
            }
            return b;
        }
        public byte[,] SuperEnc4(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);
            b = AddRoundKey(b, rk[0]);
            for (int r = 0; r < 3; r++) {
                b = SubByte(b);
                b = ShiftRows(b);
                b = MixColumns(b);
                b = AddRoundKey(b, rk[r + 1]);
            }
            b = SubByte(b);
            //b = ShiftRows(b);
            //b = MixColumns(b);
            b = AddRoundKey(b, rk[4]);
            return b;
        }
        public byte[,] SuperEnc5(byte[,] a, List<byte[,]> rk) {
            // S  = SB ◦ MC ◦ SB
            // L  = SR ◦ MC ◦ SR
            // Q  = SB ◦ MC ◦ SR
            // Q' = SR ◦ MC ◦ SB
            // SuperEnc5 =  S ◦ L ◦ S ◦ Q'
            a = AddRoundKey(a, rk[0]);

            //ShiftRows(a,0,BC);

            a = SubByte(a);
            a = MixColumns(a);              // Q'
            a = AddRoundKey(a, rk[1]);
            a = ShiftRows(a);

            a = SubByte(a);
            a = MixColumns(a);              // S
            a = AddRoundKey(a, rk[2]);
            a = SubByte(a);

            a = ShiftRows(a);
            a = MixColumns(a);              // L
            a = AddRoundKey(a, rk[3]);
            a = ShiftRows(a);

            a = SubByte(a);
            a = MixColumns(a);              // S
            a = AddRoundKey(a, rk[4]);
            a = SubByte(a);

            /*
              ShiftRows(a,0,BC);
              MixColumns(a,BC);
              */
            a = AddRoundKey(a, rk[5]);
            return a;
        }
        public byte[,] SuperEnc5Real(byte[,] a, List<byte[,]> rk) {
            // S  = SB ◦ MC ◦ SB
            // L  = SR ◦ MC ◦ SR
            // Q  = SB ◦ MC ◦ SR
            // Q' = SR ◦ MC ◦ SB
            // SuperEnc5 =  S ◦ L ◦ S ◦ Q'
            a = AddRoundKey(a, rk[0]);

            //a = ShiftRows(a);

            a = SubByte(a);
            a = MixColumns(a);
            a = AddRoundKey(a, rk[1]);
            a = ShiftRows(a);

            a = SubByte(a);
            a = MixColumns(a);
            a = AddRoundKey(a, rk[2]);
            a = SubByte(a);

            a = ShiftRows(a);
            a = MixColumns(a);
            a = AddRoundKey(a, rk[3]);
            a = ShiftRows(a);

            a = SubByte(a);
            a = MixColumns(a);
            a = AddRoundKey(a, rk[4]);
            a = SubByte(a);


            //a = ShiftRows(a);
            //a = MixColumns(a);

            a = AddRoundKey(a, rk[5]);
            return a;
        }

        // Decryption -----------------------------------------------------------------------------------------------------------
        public byte[,] Decrypt(byte[,] a, List<byte[,]> rk) {
            a = AddRoundKey(rk[10], a);
            a = InvShiftRows(a);
            a = InvSubByte(a);
            for (int i = 8; i >= 0; i--) {
                a = AddRoundKey(rk[i + 1], a);
                a = InvMixColumns(a);
                a = InvShiftRows(a);
                a = InvSubByte(a);
            }
            a = AddRoundKey(a, rk[0]);
            return a;
        }
        public byte[,] DecryptPartial(byte[,] a, List<byte[,]> rk, int rounds) {
            for (int r = 0; r < rounds; r++) {
                a = AddRoundKey(a, rk[rounds - r]);
                a = InvMixColumns(a);
                a = InvShiftRows(a);
                a = InvSubByte(a);
            }
            a = AddRoundKey(a, rk[0]);
            return a;
        }
        public byte[,] Dec(byte[,] a, List<byte[,]> rk, int rounds) { // For 3 and 4 rounds          
            byte[,] b = copy(a);
            for (int r = rounds; r > 0; r--) {
                b = AddRoundKey(b, rk[r]);
                b = InvMixColumns(b);
                b = InvShiftRows(b);
                b = InvSubByte(b);
            }
            b = AddRoundKey(b, rk[0]);
            return b;
        }
        public byte[,] SuperDec4(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);

            b = AddRoundKey(b, rk[4]);
            //b = InvMixColumns(b);
            //b = InvShiftRows(b);
            b = InvSubByte(b);

            for (int r = 3; r > 0; r--) {
                b = AddRoundKey(b, rk[r]);
                b = InvMixColumns(b);
                b = InvShiftRows(b);
                b = InvSubByte(b);
            }
            b = AddRoundKey(b, rk[0]);
            return b;
        }
        public byte[,] SuperDec5(byte[,] a, List<byte[,]> rk) {
            a = AddRoundKey(a, rk[5]);

            //InvMixColumns(a);
            //InvShiftRows(a);

            a = InvSubByte(a);
            a = AddRoundKey(a, rk[4]);  // S
            a = InvMixColumns(a);
            a = InvSubByte(a);

            a = InvShiftRows(a);
            a = AddRoundKey(a, rk[3]);  // L
            a = InvMixColumns(a);
            a = InvShiftRows(a);

            a = InvSubByte(a);
            a = AddRoundKey(a, rk[2]);  // S
            a = InvMixColumns(a);
            a = InvSubByte(a);

            a = InvShiftRows(a);
            a = AddRoundKey(a, rk[1]);  // Q'i
            a = InvMixColumns(a);
            a = InvSubByte(a);

            // InvShiftRows(a);
            a = AddRoundKey(a, rk[0]);
            return a;
        }
        public byte[,] SuperDec5Real(byte[,] a, List<byte[,]> rk) {
            a = AddRoundKey(a, rk[5]);

            a = InvMixColumns(a);
            a = InvShiftRows(a);

            a = InvSubByte(a);
            a = AddRoundKey(a, rk[4]);
            a = InvMixColumns(a);
            a = InvSubByte(a);

            a = InvShiftRows(a);
            a = AddRoundKey(a, rk[3]);
            a = InvMixColumns(a);
            a = InvShiftRows(a);

            a = InvSubByte(a);
            a = AddRoundKey(a, rk[2]);
            a = InvMixColumns(a);
            a = InvSubByte(a);

            a = InvShiftRows(a);
            a = AddRoundKey(a, rk[1]);
            a = InvMixColumns(a);
            a = InvSubByte(a);

            a = InvShiftRows(a);
            a = AddRoundKey(a, rk[0]);
            return a;
        }

        // Other stuff-----------------------------------------------------------------------------------------------------------
        public byte[,] XOR(byte[,] a, byte[,] b) {
            byte[,] c = new byte[4, 4];
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    c[i, j] = (byte)(a[i, j] ^ b[i, j]);
                }
            }
            return c;
        }
        public byte[,] XOR4(byte[,] a, byte[,] b, byte[,] c, byte[,] d) {
            return XOR(a, XOR(b, XOR(c, d)));
        }
        public byte[,] FlipMatrix(byte[,] a) {
            byte[,] b = new byte[4, 4];
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    b[i, j] = a[j, i];
                }
            }
            return b;
        }
        public byte[,] StringTo4x4ByteArray(string s) {
            byte[,] a = new byte[4, 4];
            byte[] byteValue = StringToByteArrayFastest(s);
            int counter = 0;

            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    a[j, i] = byteValue[counter];
                    counter++;
                }
            }
            return a;
        }
        public byte[] StringToByteArrayFastest(string hex) {
            if (hex.Length % 2 == 1)
                throw new Exception("The binary key cannot have an odd number of digits");

            byte[] arr = new byte[hex.Length >> 1];

            for (int i = 0; i < hex.Length >> 1; ++i) {
                arr[i] = (byte)((GetHexVal(hex[i << 1]) << 4) + (GetHexVal(hex[(i << 1) + 1])));
            }
            return arr;
        }
        public int GetHexVal(char hex) {
            int val = (int)hex;
            return val - (val < 58 ? 48 : (val < 97 ? 55 : 87));
        }
        public byte[,] copy(byte[,] x) {
            byte[,] y = new byte[4, 4];
            for (int i = 0; i < 4; i++) { //Copy x and y into a and b
                for (int j = 0; j < 4; j++) {
                    y[i, j] = x[i, j];
                }
            }
            return y;
        }
        public byte[] miniCopy(byte[] a) {
            byte[] b = new byte[4];
            for (int i = 0; i < 4; i++) {
                b[i] = a[i];
            }
            return b;
        }
        public byte RandomByte() {
            return (byte)(rand.Next(0, 256));
        }
        public byte[,] Random4x4ByteArray() {
            byte[,] randArr = new byte[4, 4];
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    randArr[i, j] = RandomByte();
                }
            }
            return randArr;
        }
        public byte[,] Zero4x4Array() {
            byte[,] zeroArr = new byte[4, 4];
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    zeroArr[i, j] = 0x00;
                }
            }
            return zeroArr;
        }


        // Super functions
        public int LinearRelation(byte[,] a, byte[,] b, byte[,] candidate_key) {
            a = AddRoundKey(a, candidate_key);
            a = SubByte(a);
            a = MixColumns(a);

            b = AddRoundKey(b, candidate_key);
            b = SubByte(b);
            b = MixColumns(b);

            if (a[2, 0] == b[2, 0]) { return 1; }
            else { return 0; }
        }
        public bool BadZeroWeightDetected(byte[,] a, byte[,] b) { // if 2 or 3 words is same in a column   /Hamming weight         
            int wt = 0;
            for (int i = 0; i < 4; i++) {
                wt = 0;
                for (int j = 0; j < 4; j++) {
                    if (a[j, i] == b[j, i])
                        wt += 1;
                }
                if (wt >= 2 && wt < 4) { //if wt is 2 or 3                    
                    return true;
                }
            }
            return false;
        }
        public int BadZeroWeightDetectedInColumn(byte[,] a, byte[,] b) { // if 2 or 3 words is same in a column            
            int wt = 0;
            for (int i = 0; i < 4; i++) {
                wt = 0;
                for (int j = 0; j < 4; j++) {
                    if (a[j, i] == b[j, i])
                        wt += 1;
                }
                if (wt >= 2 && wt < 4) { //if wt is 2 or 3                    
                    return i;
                }
            }
            return 99;
        }
        public bool BadZeroWeight(byte[,] a) {
            int wt;
            for (int i = 0; i < 4; i++) {
                wt = 0;
                for (int j = 0; j < 4; j++) {
                    if (a[j, i] == 0) { wt++; }
                }
                if (wt >= 2) { return true; }
            }
            return false;
        }

        public int NumberOfBadZeroWeightDetectedInTheColumn(byte[,] a, byte[,] b) { // if 2 or 3 words is same in a column            
            int wt = 0;
            for (int i = 0; i < 4; i++) {
                wt = 0;
                for (int j = 0; j < 4; j++) {
                    if (a[j, i] == b[j, i])
                        wt += 1;
                }
                if (wt >= 2 && wt < 4) { //if wt is 2 or 3                    
                    return wt;
                }
            }
            return wt;
        }
        public int NumberOfZeroesInColumn(byte[,] a) {
            int zeroes = 0;
            for (int i = 0; i < 4; i++) {
                if (a[i, 0] == 0) { zeroes++; }
            }
            return zeroes;
        }
        public bool HasSameDiagonal(Pair pair) {
            if (pair.p1[0, 0] == pair.p2[0, 0]) {
                if (pair.p1[1, 1] == pair.p2[1, 1]) {
                    if (pair.p1[2, 2] == pair.p2[2, 2]) {
                        if (pair.p1[3, 3] == pair.p2[3, 3]) {
                            return true;
                        }
                    }
                }
            }
            return false;
        }
        public bool Has_a_ZeroColumn(byte[,] a) {
            for (int col = 0; col < 4; col++) {
                if (a[0, col] == 0 && a[1, col] == 0 && a[2, col] == 0 && a[3, col] == 0) { return true; }
            }
            return false;
        }
        public int TheZeroColumn(byte[,] a) {
            for (int col = 0; col < 4; col++) {
                if (a[0, col] == 0 && a[1, col] == 0 && a[2, col] == 0 && a[3, col] == 0) { return col; }
            }
            return -1;
        }
        public byte[] FirstZeroColumnVector(byte[,] a) {
            byte[] v = new byte[4] { 0, 0, 0, 0 };
            for (int col = 0; col < 4; col++) {
                if (a[0, 0] == 0 && a[1, 0] == 0 && a[2, 0] == 0 && a[3, 0] == 0) { v[col] = 1; return v; }
            }
            return v;
        }
        public bool ColXisZero(byte[,] a, int col) {
            if (a[0, col] == 0 && a[1, col] == 0 && a[2, col] == 0 && a[3, col] == 0) { return true; }
            else { return false; }
        }
        public bool DiaXisZero(byte[,] a, int dia) {
            if (a[0, dia] == 0 && a[1, (dia + 1) % 4] == 0 && a[2, (dia + 2) % 4] == 0 && a[3, (dia + 3) % 4] == 0) { return true; }
            else { return false; }
        }
        public string CollisionStringInColNumber(List<string> list, int col) {
            var hashset = new HashSet<string>();
            List<string> coolList = new List<string>();
            string collision = "";
            switch (col) {
                case 0: foreach (var state in list) { coolList.Add(state.Substring(0, 8)); } break;
                case 1: foreach (var state in list) { coolList.Add(state.Substring(8, 8)); } break;
                case 2: foreach (var state in list) { coolList.Add(state.Substring(16, 8)); } break;
                case 3: foreach (var state in list) { coolList.Add(state.Substring(24, 8)); } break;
            }
            foreach (var coolString in coolList) { if (!hashset.Add(coolString)) { collision = coolString.ToString(); break; } }

            return collision;
        }
        public List<int> CollisionNumbers(List<string> list, string collision, int col) {
            List<int> numbers = new List<int>();
            var hashset = new HashSet<string>();
            List<string> coolList = new List<string>();
            switch (col) {
                case 0: foreach (var state in list) { coolList.Add(state.Substring(0, 8)); } break;
                case 1: foreach (var state in list) { coolList.Add(state.Substring(8, 8)); } break;
                case 2: foreach (var state in list) { coolList.Add(state.Substring(16, 8)); } break;
                case 3: foreach (var state in list) { coolList.Add(state.Substring(24, 8)); } break;
            }
            int num = 0;
            foreach (string s in coolList) {
                if (s == collision) {
                    numbers.Add(num);
                }
                num++;
            }
            return numbers;
        }

        //Bruteforce
        public List<byte[]> BruteRK_col_hitListCool_Parallel(byte[,] p1, byte[,] p2, byte[,] c1, byte[,] c2, int col) {
            List<byte[]> hitList = new List<byte[]>();
            byte[] mix1 = new byte[4]; byte[] mix2 = new byte[4];
            byte[] key1 = new byte[4]; byte[] key2 = new byte[4]; byte[] keyX = new byte[4];
            byte[,] SRp1 = SR(p1);
            byte[,] SRp2 = SR(p2);
            byte[] candRK = new byte[4];

            //Parallel.For(0, 256, k0 => {        });                        
            for (int k0 = 0; k0 < 256; k0++) {
                candRK[0] = (byte)k0; Console.Write(k0.ToString() + " ");
                for (int k1 = 0; k1 < 256; k1++) {
                    candRK[1] = (byte)k1;
                    for (int k2 = 0; k2 < 256; k2++) {
                        candRK[2] = (byte)k2;
                        for (int k3 = 0; k3 < 256; k3++) {
                            candRK[3] = (byte)k3;
                            for (int i = 0; i < 4; i++) {
                                mix1[i] = (byte)(c1[i, col] ^ candRK[i]);
                                mix2[i] = (byte)(c2[i, col] ^ candRK[i]);
                            }
                            for (int i = 0; i < 4; i++) {
                                key1[i] = (byte)(SRp1[i, col] ^ Boxes.si_box[(byte)(Mul(0xe, mix1[i]) ^ Mul(0xb, mix1[(i + 1) % 4]) ^ Mul(0xd, mix1[(i + 2) % 4]) ^ Mul(0x9, mix1[(i + 3) % 4]))]);
                                key2[i] = (byte)(SRp2[i, col] ^ Boxes.si_box[(byte)(Mul(0xe, mix2[i]) ^ Mul(0xb, mix2[(i + 1) % 4]) ^ Mul(0xd, mix2[(i + 2) % 4]) ^ Mul(0x9, mix2[(i + 3) % 4]))]);
                            }
                            for (int i = 0; i < 4; i++) {
                                keyX[i] = (byte)(key1[i] ^ key2[i]);
                            }
                            if (keyX[0] == 0 && keyX[1] == 0 && keyX[2] == 0 && keyX[3] == 0) {
                                hitList.Add(miniCopy(candRK)); Console.Write("Hit!: "); miniPrint(candRK);
                            }
                        }
                    }
                }
            }
            return hitList;
        }

        public List<byte[]> BruteRK_col_hitListCool(byte[,] p1, byte[,] p2, byte[,] c1, byte[,] c2, int col) {
            List<byte[]> hitList = new List<byte[]>();
            byte[] mix1 = new byte[4]; byte[] mix2 = new byte[4];
            byte[] key1 = new byte[4]; byte[] key2 = new byte[4]; byte[] keyX = new byte[4];
            byte[,] pk1 = new byte[4, 4]; byte[,] pk2 = new byte[4, 4];

            byte[,] SRp1 = SR(p1);
            byte[,] SRp2 = SR(p2);

            byte[] candRK = new byte[4];
            for (int k0 = 0; k0 < 256; k0++) {
                candRK[0] = (byte)k0; Console.Write(k0.ToString() + " ");
                for (int k1 = 0; k1 < 256; k1++) {
                    candRK[1] = (byte)k1;
                    for (int k2 = 0; k2 < 256; k2++) {
                        candRK[2] = (byte)k2;
                        for (int k3 = 0; k3 < 256; k3++) {
                            candRK[3] = (byte)k3;
                            for (int i = 0; i < 4; i++) {
                                mix1[i] = (byte)(c1[i, col] ^ candRK[i]);
                                mix2[i] = (byte)(c2[i, col] ^ candRK[i]);
                            }
                            for (int i = 0; i < 4; i++) {
                                key1[i] = (byte)(SRp1[i, col] ^ Boxes.si_box[(byte)(Mul(0xe, mix1[i]) ^ Mul(0xb, mix1[(i + 1) % 4]) ^ Mul(0xd, mix1[(i + 2) % 4]) ^ Mul(0x9, mix1[(i + 3) % 4]))]);
                                key2[i] = (byte)(SRp2[i, col] ^ Boxes.si_box[(byte)(Mul(0xe, mix2[i]) ^ Mul(0xb, mix2[(i + 1) % 4]) ^ Mul(0xd, mix2[(i + 2) % 4]) ^ Mul(0x9, mix2[(i + 3) % 4]))]);
                            }
                            for (int i = 0; i < 4; i++) {
                                keyX[i] = (byte)(key1[i] ^ key2[i]);
                            }
                            if (keyX[0] == 0 && keyX[1] == 0 && keyX[2] == 0 && keyX[3] == 0) {
                                hitList.Add(candRK); Console.WriteLine("Hit!");
                            }
                        }
                    }
                }
            }
            return hitList;
        }
        public List<byte[,]> BruteRK_col_hitList(byte[,] p1, byte[,] p2, byte[,] c1, byte[,] c2, byte[,] candRK, int col) {
            List<byte[,]> hitList = new List<byte[,]>();
            byte[,] mix1 = new byte[4, 4]; byte[,] mix2 = new byte[4, 4];
            byte[,] key1 = new byte[4, 4]; byte[,] key2 = new byte[4, 4]; byte[,] keyX = new byte[4, 4];
            byte[,] pk1 = new byte[4, 4]; byte[,] pk2 = new byte[4, 4];

            for (int k0 = 0; k0 < 256; k0++) {
                candRK[0, col] = (byte)k0; Console.WriteLine(ToString(candRK));
                for (int k1 = 0; k1 < 256; k1++) {
                    candRK[1, col] = (byte)k1;
                    for (int k2 = 0; k2 < 256; k2++) {
                        candRK[2, col] = (byte)k2;
                        for (int k3 = 0; k3 < 256; k3++) {
                            candRK[3, col] = (byte)k3;

                            for (int i = 0; i < 4; i++) {
                                for (int j = 0; j < 4; j++) {
                                    mix1[i, j] = (byte)(c1[i, j] ^ candRK[i, j]);
                                    mix2[i, j] = (byte)(c2[i, j] ^ candRK[i, j]);
                                }
                            }
                            pk1 = SBi_SRi_MCi(mix1);
                            pk2 = SBi_SRi_MCi(mix2);
                            for (int i = 0; i < 4; i++) {
                                for (int j = 0; j < 4; j++) {
                                    key1[i, j] = (byte)(p1[i, j] ^ pk1[i, j]);
                                    key2[i, j] = (byte)(p2[i, j] ^ pk2[i, j]);
                                }
                            }
                            for (int i = 0; i < 4; i++) {
                                for (int j = 0; j < 4; j++) {
                                    keyX[i, j] = (byte)(key1[i, j] ^ key2[i, j]);
                                }
                            }
                            if (DiaXisZero(keyX, col)) { hitList.Add(candRK); Console.WriteLine("Hit: " + ToString(candRK)); }

                        }
                    }
                }
            }
            return hitList;
        }
        public byte[,] BruteRK_col(byte[,] p1, byte[,] p2, byte[,] c1, byte[,] c2, byte[,] candRK, int col) {
            byte[,] mix1 = new byte[4, 4]; byte[,] mix2 = new byte[4, 4];
            byte[,] key1 = new byte[4, 4]; byte[,] key2 = new byte[4, 4]; byte[,] keyX = new byte[4, 4];

            for (int k0 = 0; k0 < 256; k0++) {
                candRK[0, col] = (byte)k0;
                for (int k1 = 0; k1 < 256; k1++) {
                    candRK[1, col] = (byte)k1; Console.WriteLine(ToString(candRK));
                    for (int k2 = 0; k2 < 256; k2++) {
                        candRK[2, col] = (byte)k2;
                        for (int k3 = 0; k3 < 256; k3++) {
                            candRK[3, col] = (byte)k3;
                            mix1 = XOR(c1, candRK);
                            mix2 = XOR(c2, candRK);

                            //SBi_SRi_MCi
                            key1 = XOR(p1, SBi_SRi_MCi(mix1));
                            key2 = XOR(p2, SBi_SRi_MCi(mix2));

                            //key1 = XOR(p1, SBi(SRi(MCi(mix1)))); //key1 = SR(key1); 
                            //key2 = XOR(p2, SBi(SRi(MCi(mix2)))); //key2 = SR(key2);

                            keyX = XOR(key1, key2);

                            //if (ColXisZero(keyX, col)) { return candRK; }
                            if (DiaXisZero(keyX, col)) { return candRK; }
                        }
                    }
                }
            }
            return candRK;
        }
        public byte[,] BruteRK_dia(byte[,] p1, byte[,] p2, byte[,] c1, byte[,] c2, byte[,] candRK, int col) {
            byte[,] mix1 = new byte[4, 4]; byte[,] mix2 = new byte[4, 4];
            byte[,] key1 = new byte[4, 4]; byte[,] key2 = new byte[4, 4]; byte[,] keyX = new byte[4, 4];
            // Bruteforce 1 column of candRK
            for (int k0 = 0; k0 < 256; k0++) {
                candRK[0, col] = (byte)k0;
                for (int k1 = 0; k1 < 256; k1++) {
                    candRK[1, col] = (byte)k1;
                    Console.WriteLine(ToString(candRK));
                    for (int k2 = 0; k2 < 256; k2++) {
                        candRK[2, col] = (byte)k2;
                        for (int k3 = 0; k3 < 256; k3++) {
                            candRK[3, col] = (byte)k3;
                            mix1 = XOR(c1, candRK);
                            mix2 = XOR(c2, candRK);

                            key1 = XOR(p1, SBi(SRi(MCi(mix1))));
                            key2 = XOR(p2, SBi(SRi(MCi(mix2))));

                            keyX = XOR(key1, key2);

                            if (DiaXisZero(keyX, col)) {
                                return candRK;
                            }
                        }
                    }
                }
            }
            return candRK;
        }

        public List<byte> CrackByte(int num, int rounds, byte[,] p1, byte[,] p2, byte[] v, byte[,] x, int i, int j) {
            List<byte> byteList = new List<byte>();
            byte[,] candK = StringTo4x4ByteArray("00000000000000000000000000000000"); List<byte[,]> candRK = new List<byte[,]>();
            byte[,] c1 = new byte[4, 4]; byte[,] c2 = new byte[4, 4];
            byte[,] p1_ = new byte[4, 4]; byte[,] p2_ = new byte[4, 4]; byte[,] c1_ = new byte[4, 4]; byte[,] c2_ = new byte[4, 4];
            byte[,] y = new byte[4, 4];
            for (int k0 = 0; k0 < 256; k0++) {
                candK[i, j] = (byte)k0; candRK = KeySchedule(candK);
                c1 = Enc(p1, candRK, rounds);
                c2 = Enc(p2, candRK, rounds);
                c1_ = PvC(c1, c2, v);
                c2_ = PvC(c2, c1, v);
                p1_ = Dec(c1_, candRK, rounds);
                p2_ = Dec(c2_, candRK, rounds);
                y = XOR(p1_, p2_); //aes.PrintMx(y);

                if (x[i, j] == y[i, j]) {
                    //Console.WriteLine(num.ToString() + "nd byte of the key is: " + candK[i, j].ToString("X").PadLeft(2, '0').ToUpper());
                    //PrintMx(y);
                    byteList.Add(candK[i, j]);
                }
            }



            return byteList;
        }

        public List<byte> CrackBytePVM(int rounds, byte[,] p1, byte[,] p2, byte[] v, byte[,] x, int i, int j) {
            List<byte> byteList = new List<byte>();
            byte[,] candK = StringTo4x4ByteArray("00000000000000000000000000000000"); List<byte[,]> candRK = new List<byte[,]>();
            byte[,] c1 = new byte[4, 4]; byte[,] c2 = new byte[4, 4];
            byte[,] p1_ = new byte[4, 4]; byte[,] p2_ = new byte[4, 4]; byte[,] c1_ = new byte[4, 4]; byte[,] c2_ = new byte[4, 4];
            byte[,] y = new byte[4, 4];
            for (int k0 = 0; k0 < 256; k0++) {
                candK[i, j] = (byte)k0; candRK = KeySchedule(candK);
                c1 = Enc(p1, candRK, rounds);
                c2 = Enc(p2, candRK, rounds);
                c1_ = PvM(c1, c2, v);
                c2_ = PvM(c2, c1, v);
                p1_ = Dec(c1_, candRK, rounds);
                p2_ = Dec(c2_, candRK, rounds);
                y = XOR(p1_, p2_);
                if (x[i, j] == y[i, j]) {
                    byteList.Add(candK[i, j]);
                }
            }
            return byteList;
        }
        public byte[,] BruteColumn(byte[,] a, int col) {
            a[3, col] += 1 % 256;
            if (a[3, col] == 0) {
                a[2, col] += 1 % 256;
                if (a[2, col] == 0) {
                    a[1, col] += 1 % 256;
                    if (a[1, col] == 0) {
                        a[0, col] += 1 % 256;
                    }
                }
            }
            return a;
        }
        public byte[,] BruteDiagonal(byte[,] a, int d) {
            a[0, (d + 0) % 4] += 1 % 256;
            if (a[0, (d + 0) % 4] == 0) {
                a[1, (d + 1) % 4] += 1 % 256;
                if (a[1, (d + 1) % 4] == 0) {
                    a[2, (d + 2) % 4] += 1 % 256;
                    if (a[2, (d + 2) % 4] == 0) {
                        a[3, (d + 3) % 4] += 1 % 256;
                    }
                }
            }
            return a;
        }
        public byte[,] Bruteforce(byte[,] a) {
            if (a[3, 3] < 255) { a[3, 3]++; }
            else if (a[3, 3] == 255) {
                a[3, 3] = 0;
                if (a[2, 3] < 255) { a[2, 3]++; }
                else if (a[2, 3] == 255) {
                    a[2, 3] = 0;
                    if (a[1, 3] < 255) { a[1, 3]++; }
                    else if (a[1, 3] == 255) {
                        a[1, 3] = 0;
                        if (a[0, 3] < 255) { a[0, 3]++; }
                        else if (a[0, 3] == 255) {
                            a[0, 3] = 0;
                            if (a[3, 2] < 255) { a[3, 2]++; }
                            else if (a[3, 2] == 255) {
                                a[3, 2] = 0;
                                if (a[2, 2] < 255) { a[2, 2]++; }
                                else if (a[2, 2] == 255) {
                                    a[2, 2] = 0;
                                    if (a[1, 2] < 255) { a[1, 2]++; }
                                    else if (a[1, 2] == 255) {
                                        a[1, 2] = 0;
                                        if (a[0, 2] < 255) { a[0, 2]++; }
                                        else if (a[0, 2] == 255) {
                                            a[0, 2] = 0;
                                            if (a[3, 1] < 255) { a[3, 1]++; }
                                            else if (a[3, 1] == 255) {
                                                a[3, 1] = 0;
                                                if (a[2, 1] < 255) { a[2, 1]++; }
                                                else if (a[2, 1] == 255) {
                                                    a[2, 1] = 0;
                                                    if (a[1, 1] < 255) { a[1, 1]++; }
                                                    else if (a[1, 1] == 255) {
                                                        a[1, 1] = 0;
                                                        if (a[0, 1] < 255) { a[0, 1]++; }
                                                        else if (a[0, 1] == 255) {
                                                            a[0, 1] = 0;
                                                            if (a[3, 0] < 255) { a[3, 0]++; }
                                                            else if (a[3, 0] == 255) {
                                                                a[3, 0] = 0;
                                                                if (a[2, 0] < 255) { a[2, 0]++; }
                                                                else if (a[2, 0] == 255) {
                                                                    a[2, 0] = 0;
                                                                    if (a[1, 0] < 255) { a[1, 0]++; }
                                                                    else if (a[1, 0] == 255) {
                                                                        a[1, 0] = 0;
                                                                        if (a[0, 0] < 255) { a[0, 0]++; }
                                                                        else if (a[0, 0] == 255) {
                                                                            a = Zero4x4Array();
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }


            }

            return a;
        }
        public byte[] BruteMini(byte[] a) {
            if (a[3] < 255) { a[3]++; }
            else if (a[3] == 255) {
                a[3] = 0;
                if (a[2] < 255) { a[2]++; }
                else if (a[2] == 255) {
                    a[2] = 0;
                    if (a[1] < 255) { a[1]++; }
                    else if (a[1] == 255) {
                        a[1] = 0;
                        if (a[0] < 255) { a[0]++; }
                        else if (a[0] == 255) {
                            a[0] = 0; a[1] = 0; a[2] = 0; a[3] = 0;
                        }
                    }
                }
            }
            return a;
        }
        public byte NextByte(ref byte b) {
            b = (byte)((b + 1) % 256);
            return b;
        }

        public byte[,] recoverFullKey3Rounds(byte[,] candidate_key, byte[,] k) {
            Console.WriteLine("Starting recover full key 3 rounds");
            List<byte[,]> rk = KeySchedule(k);

            byte[,] p1 = Zero4x4Array();
            byte[,] p2 = Zero4x4Array();

            p1[0, 0] = 1;

            /*p1 = InvShiftRows(p1);
            p1 = InvMixColumns(p1);
            p1 = InvSubByte(p1);
            p1 = AddRoundKey(p1, candidate_key);

            p2 = InvShiftRows(p2);
            p2 = InvMixColumns(p2);
            p2 = InvSubByte(p2);
            p2 = AddRoundKey(p2, candidate_key);*/

            Pair[] Pairs = new Pair[5];
            Pair P = new Pair();

            for (int i = 0; i < 5; i++) {
                P.p1 = copy(p1); //a copy
                P.p2 = copy(p2); //a copy
                Pairs[i] = P;

                p1 = Enc(p1, rk, 3);
                p2 = Enc(p2, rk, 3);
                swap(ref p1, ref p2);
                p1 = Dec(p1, rk, 3);
                p2 = Dec(p2, rk, 3);
                swap(ref p1, ref p2);
            }
            byte[] b = new byte[4];
            // b,e,9,d
            // d,b,e,9
            // 9,d,b,e
            byte[] beta = new byte[4];
            beta[0] = 0xb;
            beta[1] = 0xe;
            beta[2] = 0x9;
            beta[3] = 0xd;

            for (int z = 1; z < 4; z++) {
                int k0 = 0;
                int k1 = 0;
                int k2 = 0;
                int k3 = 0;
                int ByteFound = 0;
                int KeyRecovered = 0;

                while (k0 < 256 && KeyRecovered == 0) {
                    candidate_key[0, z] = (byte)k0;

                    k1 = 0;
                    // Find a solution for the second byte
                    while (k1 < 256 && ByteFound == 0) {
                        int con = 0;
                        for (int i = 0; i < 5; i++) {
                            b[0] = (byte)(Boxes.s_box[Pairs[i].p1[0, z] ^ k0] ^ Boxes.s_box[Pairs[i].p2[0, z] ^ k0]);
                            b[1] = (byte)(Boxes.s_box[Pairs[i].p1[1, z] ^ k1] ^ Boxes.s_box[Pairs[i].p2[1, z] ^ k1]);
                            if (Mul(Inv(beta[Mod(-(z - 1), 4)]), b[0]) == Mul(Inv(beta[Mod(-(z - 1) + 1, 4)]), b[1])) {
                                con++;
                            }
                        }
                        if (con == 5) {
                            ByteFound = 1;
                            candidate_key[1, z] = (byte)k1;
                        }
                        k1++;
                    }
                    if (ByteFound == 1) {
                        ByteFound = 0;
                    }

                    // Find a solution for the third byte
                    k2 = 0;
                    while (k2 < 256 && ByteFound == 0) {
                        int con = 0;
                        for (int i = 0; i < 5; i++) {
                            b[0] = (byte)(Boxes.s_box[Pairs[i].p1[0, z] ^ k0] ^ Boxes.s_box[Pairs[i].p2[0, z] ^ k0]);
                            b[2] = (byte)(Boxes.s_box[Pairs[i].p1[2, z] ^ k2] ^ Boxes.s_box[Pairs[i].p2[2, z] ^ k2]);
                            if (Mul(Inv(beta[Mod(-(z - 1), 4)]), b[0]) == Mul(Inv(beta[Mod(-(z - 1) + 2, 4)]), b[2])) {
                                con++;
                            }
                        }
                        if (con == 5) {
                            ByteFound = 1;
                            candidate_key[2, z] = (byte)k2;
                        }
                        k2++;
                    }
                    if (ByteFound == 1) {
                        ByteFound = 0;
                    }

                    // Find a solution for the fourth byte
                    k3 = 0;
                    while (k3 < 256 && ByteFound == 0) {
                        int con = 0;
                        // candidate_key[3][z] = k2;
                        for (int i = 0; i < 5; i++) {
                            b[0] = (byte)(Boxes.s_box[Pairs[i].p1[0, z] ^ k0] ^ Boxes.s_box[Pairs[i].p2[0, z] ^ k0]);
                            b[3] = (byte)(Boxes.s_box[Pairs[i].p1[3, z] ^ k3] ^ Boxes.s_box[Pairs[i].p2[3, z] ^ k3]);
                            if (Mul(Inv(beta[Mod(-(z - 1), 4)]), b[0]) == Mul(Inv(beta[Mod(-(z - 1) + 3, 4)]), b[3])) {
                                con++;
                            }
                        }
                        if (con == 5) {
                            ByteFound = 1;
                            candidate_key[3, z] = (byte)k3;
                        }
                        k3++;
                    }
                    if (ByteFound == 1) {
                        //Found all                        
                        Console.WriteLine(String.Format("{0}th subkey found: ", z));
                        Print(candidate_key);
                        KeyRecovered = 1;
                    }
                    k0++;
                }
            }
            Console.WriteLine("Finished recover full key");
            return candidate_key;
        }
        public byte[,] recoverFullKey5Rounds(byte[,] candidate_key, byte[,] k) {
            Console.WriteLine("Starting recover full key");
            List<byte[,]> rk = KeySchedule(k);

            byte[,] p1 = new byte[4, 4];
            byte[,] p2 = new byte[4, 4];

            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    p1[i, j] = 0;
                    p2[i, j] = 0;
                }
            }
            p1[0, 0] = 1;

            p1 = InvShiftRows(p1);
            p1 = InvMixColumns(p1);
            p1 = InvSubByte(p1);
            p1 = AddRoundKey(p1, candidate_key);

            p2 = InvShiftRows(p2);
            p2 = InvMixColumns(p2);
            p2 = InvSubByte(p2);
            p2 = AddRoundKey(p2, candidate_key);

            Pair[] Pairs = new Pair[5];
            Pair P = new Pair();

            for (int i = 0; i < 5; i++) {
                P.p1 = copy(p1); //a copy
                P.p2 = copy(p2); //a copy
                Pairs[i] = P;

                p1 = SuperEnc5(p1, rk);
                p2 = SuperEnc5(p2, rk);
                swap(ref p1, ref p2);
                p1 = SuperDec5(p1, rk);
                p2 = SuperDec5(p2, rk);
                swap(ref p1, ref p2);
            }
            byte[] b = new byte[4];
            // b,e,9,d
            // d,b,e,9
            // 9,d,b,e
            byte[] beta = new byte[4];
            beta[0] = 0xb;
            beta[1] = 0xe;
            beta[2] = 0x9;
            beta[3] = 0xd;

            for (int z = 1; z < 4; z++) {
                int k0 = 0;
                int k1 = 0;
                int k2 = 0;
                int k3 = 0;
                int ByteFound = 0;
                int KeyRecovered = 0;

                while (k0 < 256 && KeyRecovered == 0) {
                    candidate_key[0, z] = (byte)k0;

                    k1 = 0;
                    // Find a solution for the second byte
                    while (k1 < 256 && ByteFound == 0) {
                        int con = 0;
                        for (int i = 0; i < 5; i++) {
                            b[0] = (byte)(Boxes.s_box[Pairs[i].p1[0, z] ^ k0] ^ Boxes.s_box[Pairs[i].p2[0, z] ^ k0]);
                            b[1] = (byte)(Boxes.s_box[Pairs[i].p1[1, z] ^ k1] ^ Boxes.s_box[Pairs[i].p2[1, z] ^ k1]);
                            if (Mul(Inv(beta[Mod(-(z - 1), 4)]), b[0]) == Mul(Inv(beta[Mod(-(z - 1) + 1, 4)]), b[1])) {
                                con++;
                            }
                        }
                        if (con == 5) {
                            ByteFound = 1;
                            candidate_key[1, z] = (byte)k1;
                        }
                        k1++;
                    }
                    if (ByteFound == 1) {
                        ByteFound = 0;
                    }

                    // Find a solution for the third byte
                    k2 = 0;
                    while (k2 < 256 && ByteFound == 0) {
                        int con = 0;
                        for (int i = 0; i < 5; i++) {
                            b[0] = (byte)(Boxes.s_box[Pairs[i].p1[0, z] ^ k0] ^ Boxes.s_box[Pairs[i].p2[0, z] ^ k0]);
                            b[2] = (byte)(Boxes.s_box[Pairs[i].p1[2, z] ^ k2] ^ Boxes.s_box[Pairs[i].p2[2, z] ^ k2]);
                            if (Mul(Inv(beta[Mod(-(z - 1), 4)]), b[0]) == Mul(Inv(beta[Mod(-(z - 1) + 2, 4)]), b[2])) {
                                con++;
                            }
                        }
                        if (con == 5) {
                            ByteFound = 1;
                            candidate_key[2, z] = (byte)k2;
                        }
                        k2++;
                    }
                    if (ByteFound == 1) {
                        ByteFound = 0;
                    }

                    // Find a solution for the fourth byte
                    k3 = 0;
                    while (k3 < 256 && ByteFound == 0) {
                        int con = 0;
                        // candidate_key[3][z] = k2;
                        for (int i = 0; i < 5; i++) {
                            b[0] = (byte)(Boxes.s_box[Pairs[i].p1[0, z] ^ k0] ^ Boxes.s_box[Pairs[i].p2[0, z] ^ k0]);
                            b[3] = (byte)(Boxes.s_box[Pairs[i].p1[3, z] ^ k3] ^ Boxes.s_box[Pairs[i].p2[3, z] ^ k3]);
                            if (Mul(Inv(beta[Mod(-(z - 1), 4)]), b[0]) == Mul(Inv(beta[Mod(-(z - 1) + 3, 4)]), b[3])) {
                                con++;
                            }
                        }
                        if (con == 5) {
                            ByteFound = 1;
                            candidate_key[3, z] = (byte)k3;
                        }
                        k3++;
                    }
                    if (ByteFound == 1) {
                        //Found all                        
                        Console.WriteLine(String.Format("{0}th subkey found: ", z));
                        Print(candidate_key);
                        KeyRecovered = 1;
                    }
                    k0++;
                }
            }
            Console.WriteLine("Finished recover full key");
            return candidate_key;
        }
        public byte[,] recoverFullKey5RoundsReal(byte[,] candidate_key, byte[,] k) {
            Console.WriteLine("Starting recover full key");
            List<byte[,]> rk = KeySchedule(k);

            byte[,] p1 = new byte[4, 4];
            byte[,] p2 = new byte[4, 4];

            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    p1[i, j] = 0;
                    p2[i, j] = 0;
                }
            }
            p1[0, 0] = 1;

            p1 = InvShiftRows(p1);
            p1 = InvMixColumns(p1);
            p1 = InvSubByte(p1);
            p1 = AddRoundKey(p1, candidate_key);

            p2 = InvShiftRows(p2);
            p2 = InvMixColumns(p2);
            p2 = InvSubByte(p2);
            p2 = AddRoundKey(p2, candidate_key);

            Pair[] Pairs = new Pair[5];
            Pair P = new Pair();

            for (int i = 0; i < 5; i++) {
                P.p1 = copy(p1); //a copy
                P.p2 = copy(p2); //a copy
                Pairs[i] = P;

                p1 = SuperEnc5Real(p1, rk);
                p2 = SuperEnc5Real(p2, rk);
                swap(ref p1, ref p2);
                p1 = SuperDec5Real(p1, rk);
                p2 = SuperDec5Real(p2, rk);
                swap(ref p1, ref p2);
            }
            byte[] b = new byte[4];
            // b,e,9,d
            // d,b,e,9
            // 9,d,b,e
            byte[] beta = new byte[4];
            beta[0] = 0xb;
            beta[1] = 0xe;
            beta[2] = 0x9;
            beta[3] = 0xd;

            for (int z = 1; z < 4; z++) {
                int k0 = 0;
                int k1 = 0;
                int k2 = 0;
                int k3 = 0;
                int ByteFound = 0;
                int KeyRecovered = 0;

                while (k0 < 256 && KeyRecovered == 0) {
                    candidate_key[0, z] = (byte)k0;

                    k1 = 0;
                    // Find a solution for the second byte
                    while (k1 < 256 && ByteFound == 0) {
                        int con = 0;
                        for (int i = 0; i < 5; i++) {
                            b[0] = (byte)(Boxes.s_box[Pairs[i].p1[0, z] ^ k0] ^ Boxes.s_box[Pairs[i].p2[0, z] ^ k0]);
                            b[1] = (byte)(Boxes.s_box[Pairs[i].p1[1, z] ^ k1] ^ Boxes.s_box[Pairs[i].p2[1, z] ^ k1]);
                            if (Mul(Inv(beta[Mod(-(z - 1), 4)]), b[0]) == Mul(Inv(beta[Mod(-(z - 1) + 1, 4)]), b[1])) {
                                con++;
                            }
                        }
                        if (con == 5) {
                            ByteFound = 1;
                            candidate_key[1, z] = (byte)k1;
                        }
                        k1++;
                    }
                    if (ByteFound == 1) {
                        ByteFound = 0;
                    }

                    // Find a solution for the third byte
                    k2 = 0;
                    while (k2 < 256 && ByteFound == 0) {
                        int con = 0;
                        for (int i = 0; i < 5; i++) {
                            b[0] = (byte)(Boxes.s_box[Pairs[i].p1[0, z] ^ k0] ^ Boxes.s_box[Pairs[i].p2[0, z] ^ k0]);
                            b[2] = (byte)(Boxes.s_box[Pairs[i].p1[2, z] ^ k2] ^ Boxes.s_box[Pairs[i].p2[2, z] ^ k2]);
                            if (Mul(Inv(beta[Mod(-(z - 1), 4)]), b[0]) == Mul(Inv(beta[Mod(-(z - 1) + 2, 4)]), b[2])) {
                                con++;
                            }
                        }
                        if (con == 5) {
                            ByteFound = 1;
                            candidate_key[2, z] = (byte)k2;
                        }
                        k2++;
                    }
                    if (ByteFound == 1) {
                        ByteFound = 0;
                    }

                    // Find a solution for the fourth byte
                    k3 = 0;
                    while (k3 < 256 && ByteFound == 0) {
                        int con = 0;
                        // candidate_key[3][z] = k2;
                        for (int i = 0; i < 5; i++) {
                            b[0] = (byte)(Boxes.s_box[Pairs[i].p1[0, z] ^ k0] ^ Boxes.s_box[Pairs[i].p2[0, z] ^ k0]);
                            b[3] = (byte)(Boxes.s_box[Pairs[i].p1[3, z] ^ k3] ^ Boxes.s_box[Pairs[i].p2[3, z] ^ k3]);
                            if (Mul(Inv(beta[Mod(-(z - 1), 4)]), b[0]) == Mul(Inv(beta[Mod(-(z - 1) + 3, 4)]), b[3])) {
                                con++;
                            }
                        }
                        if (con == 5) {
                            ByteFound = 1;
                            candidate_key[3, z] = (byte)k3;
                        }
                        k3++;
                    }
                    if (ByteFound == 1) {
                        //Found all                        
                        Console.WriteLine(String.Format("{0}th subkey found: ", z));
                        Print(candidate_key);
                        KeyRecovered = 1;
                    }
                    k0++;
                }
            }
            Console.WriteLine("Finished recover full key");
            return candidate_key;
        }

        //ColumnExchangeDifference and Exhange operations
        public byte[,] CoD(byte[,] a, byte[,] b, byte[] v) { // (∆(α, β)_v)i  = (αi ⊕ βi)*vi
            byte[,] colExDiff = new byte[4, 4]; //ColumnExchangeDifference
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    colExDiff[i, j] = Mul((a[i, j] ^ b[i, j]), v[j]);
                }
            }
            return colExDiff;
        }
        public byte[,] DiD(byte[,] a, byte[,] b, byte[] v) { // (∆(α, β)_v)i  = (αi ⊕ βi)*vi
            byte[,] diExDiff = new byte[4, 4]; //ColumnExchangeDifference
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    diExDiff[i, (j + i) % 4] = Mul((a[i, (j + i) % 4] ^ b[i, (j + i) % 4]), v[j]);
                }
            }
            return diExDiff;
        }
        public byte[,] MiD(byte[,] a, byte[,] b, byte[] v) {
            byte[,] pvm1 = PvM(a, b, v);
            byte[,] pvm2 = PvM(b, a, v);
            return XOR(a, pvm1);
        }
        
        public byte[,] PvC(byte[,] a, byte[,] b, byte[] v) { //ColumnExchange
            return XOR(a, CoD(a, b, v));
        }
        public byte[,] PvD(byte[,] a, byte[,] b, byte[] v) { //DiagonalExchange
            return XOR(a, SRi(CoD(SR(a), SR(b), v)));
        }
        public byte[,] PvM(byte[,] a, byte[,] b, byte[] v) { //MixedExchange  
            return XOR(a, L(CoD(invL(a), invL(b), v)));
        }
        public byte[,] PvM3(byte[,] a, byte[,] b, byte[] v) { //MixedExchange  
            return XOR(a, MC(SR(SR(DiD(SRi(SRi(MCi(a))), SRi(SRi(MCi(b))), v)))));
        }

        public byte[,] PvD2(byte[,] a, byte[,] b, byte[] v) { //DiagonalExchange
            return XOR(a, DiD(a, b, v));
        }
        public byte[,] PvM2(byte[,] a, byte[,] b, byte[] v) { //MixedExchange  
            return XOR(a, MiD(a, b, v));
        }
        
        public byte[,] PvL(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) { //MixedExchange  
            //return XOR(a, SR(L(CoD(SR(invL(a)), SR(invL(b)), v))));
            //return XOR(a, SBi(L(CoD(SB(invL(a)), SB(invL(b)), v))));
            byte[,] mss1 = copy(a); mss1 = MCi(mss1); mss1 = SRi(mss1); mss1 = SBi(mss1);
            byte[,] mss2 = copy(b); mss2 = MCi(mss2); mss2 = SRi(mss2); mss2 = SBi(mss2);
            /*
            //byte[,] pvm = XOR(a, MiD(mss1, mss2, v));            
            mss1 = SR(mss1); mss1 = MC(mss1); mss1 = MC(mss1);
            mss2 = SR(mss2); mss2 = MC(mss2); mss1 = MC(mss2);
            byte[,] x = XOR(mss1, mss2);
            byte[,] L1 = XOR(a, MC(SR(SB(CoD(SBi(SRi(MCi(a))), SBi(SRi(MCi(b))), v)))));                        
            byte[,] wtf = XOR(a, SRi(MiD(SR(a), SR(b), v)));
            byte[,] bsj = XOR(a, L(CoD(invL(MiD(a, b, v)), invL(MiD(b, a, v)), v)));
            */
            byte[] vv = new byte[4] { 1, 1, 0, 0 };
            byte[,] pvd = XOR(a, MiD(a, b, v));
            byte[,] pvm = XOR(a, L(CoD(invL(a), invL(b), v)));
            byte[,] pvYolo = XOR(a, SRi(L(CoD(invL(MCi(a)), invL(MCi(b)), v))));

            byte[,] yo = XOR(a, L(CoD(invL(pvd), invL(MC(b)), v)));
            byte[,] fuck = XOR(a, MC(DiD(MCi(a), MCi(b), v))); //Wow, use v = 0101
            byte[,] asss = XOR(a, MC(DiD(MCi(invL(a)), MCi(invL(b)), v)));
            byte[,] drit = XOR(a, SR(SR(MC(MC(DiD(SRi(SRi(MCi(MCi(XOR(a, rk[4]))))), SRi(SRi(MCi(MCi(XOR(b, rk[4]))))), v))))));

            byte[,] sug1 = XOR(a, SRi(CoD(SR(a), SR(b), v)));
            byte[,] sug2 = XOR(b, SRi(CoD(SR(b), SR(a), v)));

            byte[,] hore = XOR(a, L(CoD(invL(XOR(a, invL(sug1))), invL(XOR(b, invL(sug1))), v)));

            byte[,] x = copy(a); byte[,] y = copy(b);
            x = MCi(x); x = SRi(x);
            y = MCi(y); x = SRi(y);

            byte[,] nei = PvD(XOR(a, invL(a)), L(CoD(invL(a), invL(b), v)), v);
            byte[,] yolo = MCi(MCi(SRi(CoD(XOR(a, XOR(a, b)), L(CoD(invL(a), invL(b), v)), v))));
            byte[,] yolo2 = XOR(MCi(MCi(SRi(CoD(XOR(a, XOR(a, b)), L(CoD(invL(a), invL(b), v)), v)))), L(CoD(invL(a), invL(b), v)));
            byte[,] yolo3 = XOR(a, yolo2);

            byte[,] pvOK = XOR(a, L(SB(CoD(invL(SBi(XOR(a, rk[4]))), SBi(invL(XOR(b, rk[4]))), v))));
            byte[,] pvFuck = XOR(MCiSRiSBi(XOR(a, rk[4])), SBSRMC(L(MiD(invL(MCiSRiSBi(XOR(a, rk[4]))), invL(MCiSRiSBi(XOR(b, rk[4]))), v))));
            byte[,] pvAlla = XOR(a, L(CoD(invL(XOR(a, rk[4])), invL(XOR(b, rk[4])), v)));
            byte[,] pvBlla = XOR(a, L(SB(L(CoD(invL(XOR(rk[3], SBi(invL(XOR(a, rk[4]))))), invL(XOR(rk[3], SBi(invL(XOR(b, rk[4]))))), v)))));
            //byte[,] pvClla = XOR(a, L(L(CoD(invL(XOR(rk[3], invL(XOR(a, rk[4])))), invL(XOR(rk[3], invL(XOR(b, rk[4])))), v))));

            byte[,] A = copy(a);
            A = XOR(A, rk[4]); A = MCi(A); A = SRi(A); A = SBi(A);
            A = XOR(A, rk[3]); A = MCi(A); A = SRi(A); A = SBi(A);

            byte[,] B = copy(b);
            B = XOR(B, rk[4]); B = MCi(B); B = SRi(B); B = SBi(B);
            B = XOR(B, rk[3]); B = MCi(B); B = SRi(B); B = SBi(B);

            byte[,] pvAB = PvC(A, B, v);
            A = copy(pvAB);
            //A = XOR(A, pvAB);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[3]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[4]);

            byte[,] puff = XOR(A, L(CoD(invL(A), invL(B), v)));

            return A;
        }

        public byte[,] Pv4(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            byte[,] A = copy(a);
            A = XOR(A, rk[4]); A = MCi(A); A = SRi(A); A = SBi(A);
            /*A = XOR(A, rk[3]);*/ A = MCi(A); A = SRi(A); A = SBi(A);

            byte[,] B = copy(b);
            B = XOR(B, rk[4]); B = MCi(B); B = SRi(B); B = SBi(B);
            /*B = XOR(B, rk[3]);*/ B = MCi(B); B = SRi(B); B = SBi(B);

            byte[,] pvAB = PvC(A, B, v);
            A = copy(pvAB);            
            A = SB(A); A = SR(A); A = MC(A); //A = XOR(A, rk[3]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[4]);

            byte[,] puff = XOR(A, L(CoD(invL(A), invL(B), v)));

            return A;
        }
        public byte[,] Pv42(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {            
            byte[,] A = copy(a);
            A = XOR(A, rk[4]); 
            A = MCi(A); A = SRi(A); A = SBi(A);            
            A = MCi(A); A = SRi(A); A = SBi(A);

            byte[,] B = copy(b);
            B = XOR(B, rk[4]); 
            B = MCi(B); B = SRi(B); B = SBi(B);            
            B = MCi(B); B = SRi(B); B = SBi(B);

            A = PvC(A, B, v);            
            A = SB(A); A = SR(A); A = MC(A); 
            A = SB(A); A = SR(A); A = MC(A);
            A = XOR(A, rk[4]);
            return A;
        }
        public byte[,] Pv43(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {            
            byte[,] A = copy(a);
            A = XOR(A, rk[4]);
            A = MCi(A); A = SRi(A); A = SBi(A);            

            byte[,] B = copy(b);
            B = XOR(B, rk[4]);            
            B = MCi(B); B = SRi(B); B = SBi(B);

            A = PvM(A, B, v);
            
            A = SB(A); A = SR(A); A = MC(A);
            A = XOR(A, rk[4]);
            return A;
        }
        public byte[,] Pv44(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            byte[,] A = copy(a); byte[,] B = copy(b);
            byte[,] pvmAB = PvM(A, B, v);
            byte[,] pvmBA = PvM(B, A, v);
            byte[,] x = XOR(pvmAB, pvmBA); //= 9E.....
            //byte[,] xpvm = XOR(A, x); // = B

            A = XOR(A, rk[4]);
            A = MCi(A); A = SRi(A); A = SBi(A);            
            B = XOR(B, rk[4]);
            B = MCi(B); B = SRi(B); B = SBi(B);
            
            A = PvM(A, B, v);
            
            A = SB(A); A = SR(A); A = MC(A);
            A = XOR(A, rk[4]);                        

            return A;
        }
        public byte[,] Pv45(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            byte[,] A = copy(a); byte[,] B = copy(b);
            byte[,] pvmAB = PvM(A, B, v);
            byte[,] pvmBA = PvM(B, A, v);
            byte[,] x = XOR(pvmAB, pvmBA); //= 9E.....
            //byte[,] xpvm = XOR(A, x); // = B

            // 10  20    01 11    02 32   23 33

            //XOR(a, L(CoD(invL(a), invL(b), v)));

            A = XOR(A, rk[4]);            
            A = MCi(A); A = SRi(A); A = SBi(A);
            //A[1, 0] = B[1, 0]; A[2, 0] = B[2, 0]; //A[0, 1] = 0; A[1, 1] = 0; A[0, 2] = 0; A[3, 2] = 0; A[2, 3] = 0; A[3, 3] = 0;

            B = XOR(B, rk[4]);
            B = MCi(B); B = SRi(B); B = SBi(B);
            //B[1, 0] = A[1, 0]; B[2, 0] = A[2, 0]; //B[0, 1] = 0; B[1, 1] = 0; B[0, 2] = 0; B[3, 2] = 0; B[2, 3] = 0; B[3, 3] = 0;

            byte[,] Ai = invL(A); //A[1, 0] = B[1, 0]; //A[2, 0] = B[2, 0];
            byte[,] Bi = invL(B); //B[1, 0] = A[1, 0]; //B[2, 0] = A[2, 0];

            byte[,] L1 = L(CoD(Ai, Bi, v));
            
            /*
            byte[,] cod = CoD(A, B, v);
            byte[,] did = DiD(A, B, v);
            byte[,] mid = MiD(A, B, v);
            byte[,] yoo = XOR(cod, did);
            yoo = XOR(yoo, mid);*/

            //A = copy(yoo);

            byte[,] pvm = PvM(A, B, v);
            A = XOR(A, L1);

            A = SB(A); A = SR(A); A = MC(A);
            A = XOR(A, rk[4]);



            return A;
        }
        public byte[,] Pv5(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            byte[,] A = copy(a);
            A = XOR(A, rk[5]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[4]);
            A = MCi(A); A = SRi(A); A = SBi(A);
            A = MCi(A); A = SRi(A); A = SBi(A);

            byte[,] B = copy(b);
            B = XOR(B, rk[5]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[4]);
            B = MCi(B); B = SRi(B); B = SBi(B);
            B = MCi(B); B = SRi(B); B = SBi(B);

            A = PvC(A, B, v);
            A = SB(A); A = SR(A); A = MC(A);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[4]);
            A = SB(A); A = SR(A); A = MC(A);
            A = XOR(A, rk[5]);
            return A;
        }
        public byte[,] Pv52(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            byte[,] A = copy(a);
            A = XOR(A, rk[5]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[4]);
            A = MCi(A); A = SRi(A); A = SBi(A);
            //A = MCi(A); A = SRi(A); A = SBi(A);

            byte[,] B = copy(b);
            B = XOR(B, rk[5]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[4]);
            B = MCi(B); B = SRi(B); B = SBi(B);
            //B = MCi(B); B = SRi(B); B = SBi(B);

            A = PvM(A, B, v);
            //A = SB(A); A = SR(A); A = MC(A);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[4]);
            A = SB(A); A = SR(A); A = MC(A);
            A = XOR(A, rk[5]);
            return A;
        }
        public byte[,] Pv6(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            byte[,] A = copy(a);
            A = XOR(A, rk[6]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[5]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[4]);
            A = MCi(A); A = SRi(A); A = SBi(A);
            A = MCi(A); A = SRi(A); A = SBi(A);

            byte[,] B = copy(b);
            B = XOR(B, rk[6]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[5]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[4]);
            B = MCi(B); B = SRi(B); B = SBi(B);
            B = MCi(B); B = SRi(B); B = SBi(B);

            A = PvC(A, B, v);
            A = SB(A); A = SR(A); A = MC(A);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[4]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[5]);
            A = SB(A); A = SR(A); A = MC(A);
            A = XOR(A, rk[6]);
            return A;
        }
        public byte[,] Pv7(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            byte[,] A = copy(a);
            A = XOR(A, rk[7]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[6]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[5]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[4]);
            A = MCi(A); A = SRi(A); A = SBi(A);
            A = MCi(A); A = SRi(A); A = SBi(A);

            byte[,] B = copy(b);
            B = XOR(B, rk[7]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[6]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[5]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[4]);
            B = MCi(B); B = SRi(B); B = SBi(B);
            B = MCi(B); B = SRi(B); B = SBi(B);

            A = PvC(A, B, v);
            A = SB(A); A = SR(A); A = MC(A);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[4]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[5]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[6]);
            A = SB(A); A = SR(A); A = MC(A);
            A = XOR(A, rk[7]);
            return A;
        }
        public byte[,] Pv8(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            byte[,] A = copy(a);
            A = XOR(A, rk[8]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[7]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[6]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[5]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[4]);
            A = MCi(A); A = SRi(A); A = SBi(A);
            A = MCi(A); A = SRi(A); A = SBi(A);

            byte[,] B = copy(b);
            B = XOR(B, rk[8]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[7]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[6]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[5]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[4]);
            B = MCi(B); B = SRi(B); B = SBi(B);
            B = MCi(B); B = SRi(B); B = SBi(B);

            A = PvC(A, B, v);
            A = SB(A); A = SR(A); A = MC(A);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[4]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[5]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[6]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[7]);
            A = SB(A); A = SR(A); A = MC(A);
            A = XOR(A, rk[8]);
            return A;
        }
        public byte[,] Pv9(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            byte[,] A = copy(a);
            A = XOR(A, rk[9]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[8]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[7]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[6]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[5]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[4]);
            A = MCi(A); A = SRi(A); A = SBi(A);
            A = MCi(A); A = SRi(A); A = SBi(A);

            byte[,] B = copy(b);
            B = XOR(B, rk[9]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[8]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[7]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[6]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[5]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[4]);
            B = MCi(B); B = SRi(B); B = SBi(B);
            B = MCi(B); B = SRi(B); B = SBi(B);

            A = PvC(A, B, v);
            A = SB(A); A = SR(A); A = MC(A);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[4]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[5]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[6]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[7]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[8]);
            A = SB(A); A = SR(A); A = MC(A);
            A = XOR(A, rk[9]);
            return A;
        }
        public byte[,] Pv10(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            byte[,] A = copy(a);
            A = XOR(A, rk[10]);
            /*A = MCi(A);*/ A = SRi(A); A = SBi(A); A = XOR(A, rk[9]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[8]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[7]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[6]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[5]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[4]);
            A = MCi(A); A = SRi(A); A = SBi(A);
            A = MCi(A); A = SRi(A); A = SBi(A);

            byte[,] B = copy(b);
            B = XOR(B, rk[10]);
            /*B = MCi(B);*/ B = SRi(B); B = SBi(B); B = XOR(B, rk[9]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[8]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[7]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[6]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[5]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[4]);
            B = MCi(B); B = SRi(B); B = SBi(B);
            B = MCi(B); B = SRi(B); B = SBi(B);

            A = PvC(A, B, v);
            A = SB(A); A = SR(A); A = MC(A);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[4]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[5]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[6]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[7]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[8]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[9]);
            A = SB(A); A = SR(A); //A = MC(A);
            A = XOR(A, rk[10]);
            return A;
        }
        public byte[,] Pv102(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            byte[,] A = copy(a);
            A = XOR(A, rk[10]);
            /*A = MCi(A);*/
            A = SRi(A); A = SBi(A); A = XOR(A, rk[9]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[8]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[7]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[6]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[5]);
            A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[4]);
            A = MCi(A); A = SRi(A); A = SBi(A);
            //A = MCi(A); A = SRi(A); A = SBi(A);

            byte[,] B = copy(b);
            B = XOR(B, rk[10]);
            /*B = MCi(B);*/
            B = SRi(B); B = SBi(B); B = XOR(B, rk[9]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[8]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[7]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[6]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[5]);
            B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[4]);
            B = MCi(B); B = SRi(B); B = SBi(B);
            //B = MCi(B); B = SRi(B); B = SBi(B);

            A = PvM(A, B, v);
            //A = SB(A); A = SR(A); A = MC(A);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[4]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[5]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[6]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[7]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[8]);
            A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[9]);
            A = SB(A); A = SR(A); //A = MC(A);
            A = XOR(A, rk[10]);
            return A;
        }
        public byte[,] Pv103(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            byte[,] A = copy(a); byte[,] B = copy(b);
            A = XOR(A, rk[10]);            
            A = SRi(A); A = SBi(A); A = XOR(A, rk[9]);
            B = XOR(B, rk[10]);
            B = SRi(B); B = SBi(B); B = XOR(B, rk[9]);

            for (int i = 8; i >= 4; i--) {
                A = MCi(A); A = SRi(A); A = SBi(A); A = XOR(A, rk[i]);
                B = MCi(B); B = SRi(B); B = SBi(B); B = XOR(B, rk[i]);
            }
            A = MCi(A); A = SRi(A); A = SBi(A);
            B = MCi(B); B = SRi(B); B = SBi(B);                                           
            
            A = PvM(A, B, v);
            for (int i = 4; i <= 9; i++) {
                A = SB(A); A = SR(A); A = MC(A); A = XOR(A, rk[i]);
            }            
            A = SB(A); A = SR(A); //A = MC(A);
            A = XOR(A, rk[10]);
            return A;
        }


        public byte[,] PvYo(byte[,] a, byte[,] b, byte[] v) {
            byte[,] pvm = XOR(a, L(CoD(invL(a), invL(b), v)));
            byte[,] hore = XOR(a, L(CoD(invL(XOR(a, invL(a))), invL(XOR(b, invL(b))), v)));

            return hore;
        }
        

        public byte[,] PvMMS(byte[,] a, byte[,] b, byte[] v) { //MixedExchange  
            //return XOR(a, SR(L(CoD(SR(invL(a)), SR(invL(b)), v))));
            return XOR(a, SBi(L(CoD(SB(invL(a)), SB(invL(b)), v))));
        }


        public byte[,] PvFull(byte[,] a, byte[,] b, byte[] v) { //MixedExchange  
            return XOR(a, MC(SR(CoD(SRi(MCi(a)), SRi(MCi(b)), v))));
        }
        public byte[,] PvFull2(byte[,] a, byte[,] b, byte[] v) { //MixedExchange  
            a = SBi(a);
            b = SBi(b);
            byte[,] f = MC(SR(CoD(SRi(MCi(a)), SRi(MCi(b)), v)));
            return XOR(a, f);
        }


        public byte[,] PvB(byte[,] a, byte[,] b, int i, int j) {
            byte[,] x = copy(a);
            x[i, j] = b[i, j];
            return x;
        }

        //Diagonal exchange-invariant
        public byte[] DEI(byte[,] a) {
            byte[] A = new byte[4];
            A[0] = (byte)(a[0, 0] ^ a[0, 1] ^ a[0, 2] ^ a[0, 3]);
            A[1] = (byte)(a[1, 0] ^ a[1, 1] ^ a[1, 2] ^ a[1, 3]);
            A[2] = (byte)(a[2, 0] ^ a[2, 1] ^ a[2, 2] ^ a[2, 3]);
            A[3] = (byte)(a[3, 0] ^ a[3, 1] ^ a[3, 2] ^ a[3, 3]);
            return A;
        }
        //column exchange-invariant
        //mixed exchange-invariant



        //Sub rounds

        public byte[,] Q(byte[,] a, List<byte[,]> rk) { // Q = SB ◦ MC ◦ SR
            byte[,] b = copy(a);

            b = AddRoundKey(b, rk[0]);
            b = SR(a);
            b = MC(b);
            b = AddRoundKey(b, rk[1]);
            b = SB(b);
            return b;
        }
        public byte[,] Qi(byte[,] a, List<byte[,]> rk) { // Qi = SR ◦ MC ◦ SB
            byte[,] b = copy(a);

            b = SBi(b);
            b = AddRoundKey(b, rk[1]);
            b = MCi(b);
            b = SRi(a);
            b = AddRoundKey(b, rk[0]);
            return b;
        }
        public byte[,] Q_(byte[,] a, List<byte[,]> rk) {  // Q' = SR ◦ MC ◦ SB
            byte[,] b = copy(a);

            b = AddRoundKey(b, rk[0]);
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[1]);
            b = SR(b);
            return b;
        }
        public byte[,] Q_i(byte[,] a, List<byte[,]> rk) {  // Q'i = SB ◦ MC ◦ SR
            byte[,] b = copy(a);

            b = SRi(b);
            b = AddRoundKey(b, rk[1]);
            b = MCi(b);
            b = SBi(b);
            b = AddRoundKey(b, rk[0]);
            return b;
        }

        public byte[,] L(byte[,] a) { // L = MC ◦ SR(a)
            byte[,] l = copy(a);
            l = ShiftRows(l);
            l = MixColumns(l);
            return l;
        }
        public byte[,] L_o(byte[,] a, List<byte[,]> rk) { // L = MC ◦ SR(a)
            byte[,] l = copy(a);
            l = ShiftRows(l);
            l = MixColumns(l);
            l = AddRoundKey(l, rk[1]);
            return l;
        }
        public byte[,] invL(byte[,] a) { // L^-1 = SR^-1 ◦ MC^-1(a)
            byte[,] l = copy(a);
            l = InvMixColumns(l);
            l = InvShiftRows(l);
            return l;
        }
        public byte[,] R(byte[,] a, List<byte[,]> rk, int rounds) {
            byte[,] b = copy(a);
            b = AddRoundKey(b, rk[0]); //must be commenctet out for oneLineBreaker to work
            for (int r = 0; r < rounds; r++) {
                b = SubByte(b);
                b = ShiftRows(b);
                b = MixColumns(b);
                b = AddRoundKey(b, rk[r + 1]);
            }
            return b;
        }
        public byte[,] Ri(byte[,] a, List<byte[,]> rk, int rounds) {
            byte[,] b = copy(a);
            for (int r = rounds; r > 0; r--) {
                b = AddRoundKey(b, rk[r]);
                b = InvMixColumns(b);
                b = InvShiftRows(b);
                b = InvSubByte(b);
            }
            b = AddRoundKey(b, rk[0]);
            return b;
        }
        public byte[,] R_no_K(byte[,] a, int rounds) {
            byte[,] b = copy(a);
            //b = AddRoundKey(b, rk[0]);
            for (int r = 0; r < rounds; r++) {
                b = SubByte(b);
                b = ShiftRows(b);
                b = MixColumns(b);
                //b = AddRoundKey(b, rk[r + 1]);
            }
            return b;
        }
        public byte[,] Ri_no_K(byte[,] a, int rounds) {
            byte[,] b = copy(a);
            for (int r = rounds; r > 0; r--) {
                //b = AddRoundKey(b, rk[r]);
                b = InvMixColumns(b);
                b = InvShiftRows(b);
                b = InvSubByte(b);
            }
            //b = AddRoundKey(b, rk[0]);
            return b;
        }

        public int NumberOfNonZeroColumns(byte[,] a) { //numberOfNonZeroColumns      Endre denne sånn at den returnerer en vektor på 16 bytes
            int numberOfNonZeroColumns = 0;
            for (int j = 0; j < 4; j++) {
                if (a[0, j] != 0 || a[1, j] != 0 || a[2, j] != 0 || a[3, j] != 0) {
                    numberOfNonZeroColumns++;
                }
            }
            return numberOfNonZeroColumns;
        }
        public byte[] v(byte[,] a) {
            byte[] zeroDifferencePattern = new byte[16]; //if ai == 0  -->  vi = 1
            int count = 0;
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    if (a[j, i] == 0) { zeroDifferencePattern[count] = 1; }
                    else { zeroDifferencePattern[count] = 0; }
                    count++;
                }
            }
            return zeroDifferencePattern;
        }
        public int wt(byte[,] a) {
            int hammingWeight = 0;
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    if (a[i, j] != 0) { hammingWeight++; }
                }
            }
            return hammingWeight;
        }
        public int Wt(byte[,] a) { //numberOfNonZeroColumns      Endre denne sånn at den returnerer en vektor på 16 bytes
            int numberOfNonZeroColumns = 0;
            for (int j = 0; j < 4; j++) {
                if (a[0, j] == 0 || a[1, j] == 0 || a[2, j] == 0 || a[3, j] == 0) {
                    numberOfNonZeroColumns++;
                }
            }
            return numberOfNonZeroColumns;
        }
        public bool SameV(byte[] v1, byte[] v2) {
            for (int i = 0; i < 16; i++) {
                if (v1[i] != v2[i]) { return false; }
            }
            return true;
        }
        public bool IsTheSame(byte[,] a, byte[,] b) {
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    if (a[i, j] != b[i, j]) { return false; }
                }
            }
            return true;
        }

        /*
             S  = SB ◦ MC ◦ SB
             L  = SR ◦ MC ◦ SR
             Q  = SB ◦ MC ◦ SR
             Q' = SR ◦ MC ◦ SB
        */
        // R rounds
        public byte[,] R2(byte[,] a, List<byte[,]> rk) {    // R2  = (AK ◦ MC ◦ SR ◦ SB) ◦ (AK ◦ MC ◦ SR ◦ SB)            
            byte[,] b = copy(a);
            // R2 = MC ◦ SR ◦ (SB ◦ MC ◦ SB) ◦ SR
            // R2 = MC ◦ SR ◦ S ◦ SR
            // R2 = MC ◦ SR ◦ S
            b = AddRoundKey(b, rk[0]);

            b = SR(b); // comment out

            // S = (SB ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[1]);
            b = SB(b);

            b = SR(b);
            b = MC(b);

            b = AddRoundKey(b, rk[2]);
            return b;
        }
        public byte[,] R2i(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);

            b = AddRoundKey(b, rk[2]);
            b = MCi(b);
            b = SRi(b);

            b = SBi(b);
            b = AddRoundKey(b, rk[1]);
            b = MCi(b);
            b = SBi(b);

            b = SRi(b);

            b = AddRoundKey(b, rk[0]);
            return b;
        }
        public byte[,] R2_(byte[,] a, List<byte[,]> rk) {    // R2  = (AK ◦ MC ◦ SR ◦ SB) ◦ (AK ◦ MC ◦ SR ◦ SB)            
            byte[,] b = copy(a);
            // R2 = MC ◦ SR ◦ (SB ◦ MC ◦ SB) ◦ SR
            // R2 = MC ◦ SR ◦ S ◦ SR
            // R2 = MC ◦ SR ◦ S
            b = AddRoundKey(b, rk[0]);

            //b = SR(b); // comment out

            // S = (SB ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[1]);
            b = SB(b);

            b = SR(b);
            b = MC(b);

            b = AddRoundKey(b, rk[2]);
            return b;
        }
        public byte[,] R2_i(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);

            b = AddRoundKey(b, rk[2]);
            b = MCi(b);
            b = SRi(b);

            b = SBi(b);
            b = AddRoundKey(b, rk[1]);
            b = MCi(b);
            b = SBi(b);

            //b = SRi(b); // comment out

            b = AddRoundKey(b, rk[0]);
            return b;
        }

        public byte[,] R3(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);                        //                          Q        ◦ S               Q        ◦       S
            // R3 = Q ◦ S    =    (SB ◦ MC ◦ SR) ◦ S    =    (SB ◦ MC ◦ SR) ◦ (SB ◦ MC ◦ SB) 
            // R3 = MC ◦ SR ◦ Q ◦ S ◦ SR
            b = AddRoundKey(b, rk[0]);  // 1

            b = SR(b); // comment out   // 2

            // S = (SB ◦ MC ◦ SB)
            b = SB(b);                  // 3
            b = MC(b);                  // 4
            b = AddRoundKey(b, rk[1]);  // 5
            b = SB(b);                  // 6


            // Q = SB ◦ MC ◦ SR
            b = SR(b);                  // 7
            b = MC(b);                  // 8
            b = AddRoundKey(b, rk[2]);  // 9
            b = SB(b);                  // 10

            b = SR(b); // comment out   // 11
            b = MC(b); // comment out   // 12

            b = AddRoundKey(b, rk[3]);  // 13

            return b;
        }
        public byte[,] R3i(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);

            b = AddRoundKey(b, rk[3]);  // 13

            b = MCi(b); // comment out   // 12
            b = SRi(b); // comment out

            b = SBi(b);
            b = AddRoundKey(b, rk[2]);
            b = MCi(b);
            b = SRi(b);
            // Qi = SRi ◦ MCi ◦ SBi

            b = SBi(b);
            b = AddRoundKey(b, rk[1]);
            b = MCi(b);
            b = SBi(b);
            // Si = (SBi ◦ MCi ◦ SBi)

            b = SRi(b); // comment out

            b = AddRoundKey(b, rk[0]);

            return b;
        }
        public byte[,] R3_(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);                        //                          Q        ◦ S               Q        ◦       S
            // R3 = Q ◦ S    =    (SB ◦ MC ◦ SR) ◦ S    =    (SB ◦ MC ◦ SR) ◦ (SB ◦ MC ◦ SB) 
            // R3 = MC ◦ SR ◦ Q ◦ S ◦ SR
            b = AddRoundKey(b, rk[0]);  // 1

            //b = SR(b); // comment out   // 2

            // S = (SB ◦ MC ◦ SB)
            b = SB(b);                  // 3
            b = MC(b);                  // 4
            b = AddRoundKey(b, rk[1]);  // 5
            b = SB(b);                  // 6


            // Q = SB ◦ MC ◦ SR
            b = SR(b);                  // 7
            b = MC(b);                  // 8
            b = AddRoundKey(b, rk[2]);  // 9
            b = SB(b);                  // 10

            //b = SR(b); // comment out   // 11
            //b = MC(b); // comment out   // 12

            b = AddRoundKey(b, rk[3]);  // 13

            return b;
        }
        public byte[,] R3_i(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);

            b = AddRoundKey(b, rk[3]);  // 13

            //b = MCi(b); // comment out   // 12
            //b = SRi(b); // comment out

            b = SBi(b);
            b = AddRoundKey(b, rk[2]);
            b = MCi(b);
            b = SRi(b);
            // Qi = SRi ◦ MCi ◦ SBi

            b = SBi(b);
            b = AddRoundKey(b, rk[1]);
            b = MCi(b);
            b = SBi(b);
            // Si = (SBi ◦ MCi ◦ SBi)

            //b = SRi(b); // comment out

            b = AddRoundKey(b, rk[0]);

            return b;
        }

        public byte[,] R4(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);
            /*
                R4 Full
                R4 = (ARK) ◦ MC ◦ SR ◦ S(ARK) ◦ L(ARK) ◦ S(ARK) ◦ SR  ◦ (AK)

                R4 Full without Key
                R4 = MC ◦ SR ◦ S ◦ L ◦ S ◦ SR

                R4  minus the linear layer before the first s-box layer and after the last s-box layer
                R4 = S ◦ L ◦ S
             */
            b = AddRoundKey(b, rk[0]);

            b = SR(b); // comment out

            // S  = (SB ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[1]);
            b = SB(b);

            // L  = (SR ◦ MC ◦ SR)
            b = SR(b);
            b = MC(b);
            b = AddRoundKey(b, rk[2]);
            b = SR(b);

            // S  = (SB ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[3]);
            b = SB(b);

            b = SR(b); // comment out
            b = MC(b); // comment out

            b = AddRoundKey(b, rk[4]);

            return b;
        }
        public byte[,] R4i(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);

            b = AddRoundKey(b, rk[4]);

            b = MCi(b); // comment out
            b = SRi(b); // comment out

            // Si  = (SBi ◦ MCi ◦ SBi)
            b = SBi(b);
            b = AddRoundKey(b, rk[3]);
            b = MCi(b);
            b = SBi(b);

            // Li  = (SRi ◦ MCi ◦ SRi)
            b = SRi(b);
            b = AddRoundKey(b, rk[2]);
            b = MCi(b);
            b = SRi(b);

            // Si  = (SBi ◦ MCi ◦ SBi)
            b = SBi(b);
            b = AddRoundKey(b, rk[1]);
            b = MCi(b);
            b = SBi(b);

            b = SRi(b); // comment out

            b = AddRoundKey(b, rk[0]);

            return b;
        }
        public byte[,] R4_(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);
            /*
                R4 Full
                R4 = (ARK) ◦ MC ◦ SR ◦ S(ARK) ◦ L(ARK) ◦ S(ARK) ◦ SR  ◦ (AK)

                R4 Full without Key
                R4 = MC ◦ SR ◦ S ◦ L ◦ S ◦ SR

                R4  minus the linear layer before the first s-box layer and after the last s-box layer
                R4 = S ◦ L ◦ S
             */
            b = AddRoundKey(b, rk[0]);

            //b = SR(b); // comment out

            // S  = (SB ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[1]);
            b = SB(b);

            // L  = (SR ◦ MC ◦ SR)
            b = SR(b);
            b = MC(b);
            b = AddRoundKey(b, rk[2]);
            b = SR(b);

            // S  = (SB ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[3]);
            b = SB(b);

            //b = SR(b); // comment out
            //b = MC(b); // comment out

            b = AddRoundKey(b, rk[4]);

            return b;
        }
        public byte[,] R4_i(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);

            b = AddRoundKey(b, rk[4]);

            //b = MCi(b); // comment out
            //b = SRi(b); // comment out

            // Si  = (SBi ◦ MCi ◦ SBi)
            b = SBi(b);
            b = AddRoundKey(b, rk[3]);
            b = MCi(b);
            b = SBi(b);

            // Li  = (SRi ◦ MCi ◦ SRi)
            b = SRi(b);
            b = AddRoundKey(b, rk[2]);
            b = MCi(b);
            b = SRi(b);

            // Si  = (SBi ◦ MCi ◦ SBi)
            b = SBi(b);
            b = AddRoundKey(b, rk[1]);
            b = MCi(b);
            b = SBi(b);

            //b = SRi(b); // comment out

            b = AddRoundKey(b, rk[0]);

            return b;
        }

        public byte[,] R5(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);
            // S  = SB ◦ MC ◦ SB
            // L  = SR ◦ MC ◦ SR
            // Q  = SB ◦ MC ◦ SR
            // Q' = SR ◦ MC ◦ SB
            /*
             R5 = S ◦ L ◦ S ◦ (SR ◦ MC ◦ SB)
                = S ◦ L ◦ S ◦      Q'
                = R4 ◦ Q'
             */
            b = AddRoundKey(b, rk[0]);

            b = SR(b); // comment out

            // Q' = (SR ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[1]);
            b = SR(b);

            // S = (SB ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[2]);
            b = SB(b);

            // L  = (SR ◦ MC ◦ SR)
            b = SR(b);
            b = MC(b);
            b = AddRoundKey(b, rk[3]);
            b = SR(b);

            // S = (SB ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[4]);
            b = SB(b);


            b = SR(b); // comment out
            b = MC(b); // comment out

            b = AddRoundKey(b, rk[5]);
            return b;
        }
        public byte[,] R5i(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);

            b = AddRoundKey(b, rk[5]);

            b = MCi(b);
            b = SRi(b);

            b = SBi(b);
            b = AddRoundKey(b, rk[4]);
            b = MCi(b);
            b = SBi(b);

            b = SRi(b);
            b = AddRoundKey(b, rk[3]);
            b = MCi(b);
            b = SRi(b);

            b = SBi(b);
            b = AddRoundKey(b, rk[2]);
            b = MCi(b);
            b = SBi(b);

            b = SRi(b);
            b = AddRoundKey(b, rk[1]);
            b = MCi(b);
            b = SBi(b);

            b = SRi(b);
            b = AddRoundKey(b, rk[0]);
            return b;
        }
        public byte[,] R5_(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);
            // S  = SB ◦ MC ◦ SB
            // L  = SR ◦ MC ◦ SR
            // Q  = SB ◦ MC ◦ SR
            // Q' = SR ◦ MC ◦ SB
            /*
             R5 = S ◦ L ◦ S ◦ (SR ◦ MC ◦ SB)
                = S ◦ L ◦ S ◦      Q'
                = R4 ◦ Q'
             */
            b = AddRoundKey(b, rk[0]);

            //b = SR(b); // comment out

            // Q' = (SR ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[1]);
            b = SR(b);

            // S = (SB ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[2]);
            b = SB(b);

            // L  = (SR ◦ MC ◦ SR)
            b = SR(b);
            b = MC(b);
            b = AddRoundKey(b, rk[3]);
            b = SR(b);

            // S = (SB ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[4]);
            b = SB(b);


            //b = SR(b); // comment out
            //b = MC(b); // comment out

            b = AddRoundKey(b, rk[5]);
            return b;
        }
        public byte[,] R5_i(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);

            b = AddRoundKey(b, rk[5]);

            //b = MCi(b); // comment out
            //b = SRi(b); // comment out

            b = SBi(b);
            b = AddRoundKey(b, rk[4]);
            b = MCi(b);
            b = SBi(b);

            b = SRi(b);
            b = AddRoundKey(b, rk[3]);
            b = MCi(b);
            b = SRi(b);

            b = SBi(b);
            b = AddRoundKey(b, rk[2]);
            b = MCi(b);
            b = SBi(b);

            b = SRi(b);
            b = AddRoundKey(b, rk[1]);
            b = MCi(b);
            b = SBi(b);

            //b = SRi(b); // comment out

            b = AddRoundKey(b, rk[0]);
            return b;
        }

        public byte[,] R6(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);
            //R6 = MC ◦ SR ◦ S ◦ L ◦ S ◦ L ◦ S ◦ SR
            //   =           S ◦ L ◦ S ◦ L ◦ S
            // S  = SB ◦ MC ◦ SB
            // L  = SR ◦ MC ◦ SR
            // Q  = SB ◦ MC ◦ SR
            // Q' = SR ◦ MC ◦ SB
            b = AddRoundKey(b, rk[0]);

            b = SR(b); // comment out

            // S  = (SB ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[1]);
            b = SB(b);

            // L  = (SR ◦ MC ◦ SR)
            b = SR(b);
            b = MC(b);
            b = AddRoundKey(b, rk[2]);
            b = SR(b);

            // S  = (SB ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[3]);
            b = SB(b);

            // L  = (SR ◦ MC ◦ SR)
            b = SR(b);
            b = MC(b);
            b = AddRoundKey(b, rk[4]);
            b = SR(b);

            // S  = (SB ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[5]);
            b = SB(b);

            b = SR(b); // comment out
            b = MC(b); // comment out

            b = AddRoundKey(b, rk[6]);

            return b;
        }
        public byte[,] R6i(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);

            b = AddRoundKey(b, rk[6]);

            b = MCi(b); // comment out
            b = SRi(b); // comment out

            // Si  = (SBi ◦ MCi ◦ SBi)
            b = SBi(b);
            b = AddRoundKey(b, rk[5]);
            b = MCi(b);
            b = SBi(b);

            // Li  = (SRi ◦ MCi ◦ SRi)
            b = SRi(b);
            b = AddRoundKey(b, rk[4]);
            b = MCi(b);
            b = SRi(b);

            // Si  = (SBi ◦ MCi ◦ SBi)
            b = SBi(b);
            b = AddRoundKey(b, rk[3]);
            b = MCi(b);
            b = SBi(b);

            // Li  = (SRi ◦ MCi ◦ SRi)
            b = SRi(b);
            b = AddRoundKey(b, rk[2]);
            b = MCi(b);
            b = SRi(b);

            // Si  = (SBi ◦ MCi ◦ SBi)
            b = SBi(b);
            b = AddRoundKey(b, rk[1]);
            b = MCi(b);
            b = SBi(b);

            b = SRi(b); // comment out

            b = AddRoundKey(b, rk[0]);

            return b;
        }
        public byte[,] R6_(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);
            //R6 = MC ◦ SR ◦ S ◦ L ◦ S ◦ L ◦ S ◦ SR
            //   =           S ◦ L ◦ S ◦ L ◦ S
            // S  = SB ◦ MC ◦ SB
            // L  = SR ◦ MC ◦ SR
            // Q  = SB ◦ MC ◦ SR
            // Q' = SR ◦ MC ◦ SB
            b = AddRoundKey(b, rk[0]);

            //b = SR(b); // comment out

            // S  = (SB ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[1]);
            b = SB(b);

            // L  = (SR ◦ MC ◦ SR)
            b = SR(b);
            b = MC(b);
            b = AddRoundKey(b, rk[2]);
            b = SR(b);

            // S  = (SB ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[3]);
            b = SB(b);

            // L  = (SR ◦ MC ◦ SR)
            b = SR(b);
            b = MC(b);
            b = AddRoundKey(b, rk[4]);
            b = SR(b);

            // S  = (SB ◦ MC ◦ SB)
            b = SB(b);
            b = MC(b);
            b = AddRoundKey(b, rk[5]);
            b = SB(b);

            //b = SR(b); // comment out
            //b = MC(b); // comment out

            b = AddRoundKey(b, rk[6]);

            return b;
        }
        public byte[,] R6_i(byte[,] a, List<byte[,]> rk) {
            byte[,] b = copy(a);

            b = AddRoundKey(b, rk[6]);

            //b = MCi(b); // comment out
            //b = SRi(b); // comment out

            // Si  = (SBi ◦ MCi ◦ SBi)
            b = SBi(b);
            b = AddRoundKey(b, rk[5]);
            b = MCi(b);
            b = SBi(b);

            // Li  = (SRi ◦ MCi ◦ SRi)
            b = SRi(b);
            b = AddRoundKey(b, rk[4]);
            b = MCi(b);
            b = SRi(b);

            // Si  = (SBi ◦ MCi ◦ SBi)
            b = SBi(b);
            b = AddRoundKey(b, rk[3]);
            b = MCi(b);
            b = SBi(b);

            // Li  = (SRi ◦ MCi ◦ SRi)
            b = SRi(b);
            b = AddRoundKey(b, rk[2]);
            b = MCi(b);
            b = SRi(b);

            // Si  = (SBi ◦ MCi ◦ SBi)
            b = SBi(b);
            b = AddRoundKey(b, rk[1]);
            b = MCi(b);
            b = SBi(b);

            //b = SRi(b); // comment out

            b = AddRoundKey(b, rk[0]);

            return b;
        }

        // G rounds
        public byte[,] G2(byte[,] a) { // G2 = S ◦ L ◦ S
            byte[,] b = copy(a);
            b = SubByte(b);
            b = L(b);
            b = SubByte(b);
            return b;
        }
        public byte[,] G2i(byte[,] a) { // G2 = S ◦ L ◦ S
            byte[,] b = copy(a);
            b = InvSubByte(b);
            b = invL(b);
            b = InvSubByte(b);
            return b;
        }
        public byte[,] G3(byte[,] a) { // G2 = S ◦ L ◦ S ◦ L ◦ S
            byte[,] b = copy(a);
            b = SubByte(b);
            b = L(b);
            b = SubByte(b);
            b = L(b);
            b = SubByte(b);
            return b;
        }
        public byte[,] G3i(byte[,] a) { // G2 = S ◦ L ◦ S ◦ L ◦ S
            byte[,] b = copy(a);
            b = InvSubByte(b);
            b = invL(b);
            b = InvSubByte(b);
            b = invL(b);
            b = InvSubByte(b);
            return b;
        }

        //R1 functions
        public byte[,] R_pvcAB(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            return R(PvC(a, b, v), rk, 1);
        }
        public byte[,] pvm_RaRb(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            return PvM(R(a, rk, 1), R(b, rk, 1), v);
        }
        public byte[,] RaXL_CoD_invLRAinvLRB(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            return XOR(R(a, rk, 1), L(CoD(invL(R(a, rk, 1)), invL(R(b, rk, 1)), v)));//// R(α) ⊕ L(∆_v^(L^(-1) (R(α)),L^(-1) (R(β))  ))
        }
        public byte[,] L_oSaXcodSaSb(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            return L_o(XOR(SubByte(a), CoD(SubByte(a), SubByte(b), v)), rk);//L ◦ (SB(α) ⊕ ∆v SB(α),SB(β))
        }
        public byte[,] L_oSa_aXcodAB(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            return L_o(SubByte(XOR(a, CoD(a, b, v))), rk); //=L◦SB(α ⊕ ∆_v^(α,β ))
        }

        //R2 functions
        public byte[,] R2pvdAB(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            return R(PvD(a, b, v), rk, 2);
        }
        public byte[,] PvmR2aR2b(byte[,] a, byte[,] b, byte[] v, List<byte[,]> rk) {
            return PvM(R(a, rk, 2), R(b, rk, 2), v);
        }


        public Pair copyPair(Pair pair1) {
            Pair pair2 = new Pair();
            pair2.p1 = pair1.p1;
            pair2.p2 = pair1.p2;
            pair2.c1 = pair1.c1;
            pair2.c1 = pair1.c1;
            return pair2;
        }
        public Pair SwapDiagonals(Pair pair) {
            byte[,] tmp1 = new byte[4, 4];
            byte[,] tmp2 = new byte[4, 4];
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    tmp1[i, j] = pair.p1[i, j];
                    tmp2[i, j] = pair.p2[i, j];
                }
            }
            pair.p1[0, 0] = tmp2[0, 0];
            pair.p1[1, 1] = tmp2[1, 1];
            pair.p1[2, 2] = tmp2[2, 2];
            pair.p1[3, 3] = tmp2[3, 3];

            pair.p2[0, 0] = tmp1[0, 0];
            pair.p2[1, 1] = tmp1[1, 1];
            pair.p2[2, 2] = tmp1[2, 2];
            pair.p2[3, 3] = tmp1[3, 3];

            return pair;
        }
        public Pair SwapColumns(Pair pair, int col) {
            byte[,] tmp1 = new byte[4, 4];
            byte[,] tmp2 = new byte[4, 4];
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    tmp1[i, j] = pair.p1[i, j];
                    tmp2[i, j] = pair.p2[i, j];
                }
            }
            pair.p1[0, col] = tmp2[0, col];
            pair.p1[1, col] = tmp2[1, col];
            pair.p1[2, col] = tmp2[2, col];
            pair.p1[3, col] = tmp2[3, col];

            pair.p2[0, col] = tmp1[0, col];
            pair.p2[1, col] = tmp1[1, col];
            pair.p2[2, col] = tmp1[2, col];
            pair.p2[3, col] = tmp1[3, col];

            return pair;
        }
        public Pair SwapRows(Pair pair, int row) {
            byte[,] tmp1 = new byte[4, 4];
            byte[,] tmp2 = new byte[4, 4];
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    tmp1[i, j] = pair.p1[i, j];
                    tmp2[i, j] = pair.p2[i, j];
                }
            }
            pair.p1[row, 0] = tmp2[row, 0];
            pair.p1[row, 1] = tmp2[row, 1];
            pair.p1[row, 2] = tmp2[row, 2];
            pair.p1[row, 3] = tmp2[row, 3];

            pair.p2[row, 0] = tmp1[row, 0];
            pair.p2[row, 1] = tmp1[row, 1];
            pair.p2[row, 2] = tmp1[row, 2];
            pair.p2[row, 3] = tmp1[row, 3];

            return pair;
        }
        public Pair SimpleSwapPair(Pair pair) { //SimpleSWAP, Swaps the first word(column) where texts are different
            byte[,] a = new byte[4, 4];
            byte[,] b = new byte[4, 4];

            for (int i = 0; i < 4; i++) { //Copy x and y into a and b
                for (int j = 0; j < 4; j++) {
                    a[i, j] = pair.p1[i, j];
                    b[i, j] = pair.p2[i, j];
                }
            }
            int z = -1;
            // Pick the first word where texts are different
            for (int i = 0; i < 4; i++) { // columns 0, 1, 2, 3
                for (int j = 0; j < 4; j++) {
                    if (pair.p1[j, i] != pair.p2[j, i]) { //Goes down each column of x and y, trying to find a difference here
                        z = i; //f.eks difference in columns 2 ,  i = 2
                        break;
                    }
                }
                if (z != -1) {
                    break; // Now we have a difference in z = i, so stop this loop and continue
                }
            }
            if (z != -1) { //if any difference
                for (int i = 0; i < 4; i++) {
                    pair.p1[i, z] = b[i, z];
                    pair.p2[i, z] = a[i, z];
                }
            }
            return pair;
        }
        public Pair Enc2Pairs(Pair pair, List<byte[,]> rk) {
            pair.p1 = Encrypt(pair.p1, rk);
            pair.p2 = Encrypt(pair.p2, rk);
            return pair;
        }
        public Pair Dec2Pairs(Pair pair, List<byte[,]> rk) {
            pair.p1 = Decrypt(pair.p1, rk);
            pair.p2 = Decrypt(pair.p2, rk);
            return pair;
        }
        public Pair EncSwapDec(Pair pair, List<byte[,]> rk) {
            pair = Enc2Pairs(pair, rk);
            pair = SwapDiagonals(pair);
            pair = Dec2Pairs(pair, rk);
            return pair;
        }

        public int NuberOfSameBytes(byte[,] a, byte[,] b) {
            int counter = 0;
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    if (a[i, j] == b[i, j]) {
                        counter++;
                    }
                }
            }
            return counter;
        }

        //SwapFunctions
        public void swap(ref byte[,] x, ref byte[,] y) { //SimpleSWAP, Swaps the first word(column) where texts are different
            byte[,] a = new byte[4, 4];
            byte[,] b = new byte[4, 4];

            for (int i = 0; i < 4; i++) { //Copy x and y into a and b
                for (int j = 0; j < 4; j++) {
                    a[i, j] = x[i, j];
                    b[i, j] = y[i, j];
                }
            }
            int z = -1;
            // Pick the first word where texts are different
            for (int i = 0; i < 4; i++) { // columns 0, 1, 2, 3
                for (int j = 0; j < 4; j++) {
                    if (x[j, i] != y[j, i]) { //Goes down each column of x and y, trying to find a difference here
                        z = i; //f.eks difference in columns 2 ,  i = 2
                        break;
                    }
                }
                if (z != -1) {
                    break; // Now we have a difference in z = i, so stop this loop and continue
                }
            }
            if (z != -1) { //if any difference
                for (int i = 0; i < 4; i++) {
                    x[i, z] = b[i, z];
                    y[i, z] = a[i, z];
                }
            }
        }
        public byte[,] simpleSwap(byte[,] x, byte[,] y) { //SimpleSWAP, Swaps the first word(column) where texts are different
            byte[,] a = copy(x);
            byte[,] b = copy(y);

            int z = -1;
            // Pick the first word where texts are different
            for (int i = 0; i < 4; i++) { // columns 0, 1, 2, 3
                for (int j = 0; j < 4; j++) {
                    if (a[j, i] != b[j, i]) { //Goes down each column of x and y, trying to find a difference here
                        z = i; //f.eks difference in columns 2 ,  i = 2
                        break;
                    }
                }
                if (z != -1) {
                    break; // Now we have a difference in z = i, so stop this loop and continue
                }
            }
            if (z != -1) { //if any difference
                for (int i = 0; i < 4; i++) {
                    a[i, z] = y[i, z];
                    b[i, z] = x[i, z];
                }
            }
            return a;
        }
        public byte[,] swapC0(byte[,] a, byte[,] b) { //c'0 = MC * SR( pv(invSR * invMC(c0), invSR * invMC(c1)) )                       
            return MC(SR(simpleSwap((SRi(MCi(a))), (SRi(MCi(b))))));
        }
        public byte[,] swapC1(byte[,] a, byte[,] b) { //c'1 = MC * SR( pv(invSR * invMC(c1), invSR * invMC(c0)) )            
            return MC(SR(simpleSwap((SRi(MCi(b))), (SRi(MCi(a))))));
        }
        public byte[,] swapP1(byte[,] a, byte[,] b) { //p'0 = invSR( pv(SR(p0), SR(p1)) )            
            return InvShiftRows(simpleSwap(ShiftRows(a), ShiftRows(b)));
        }
        public byte[,] swapP2(byte[,] a, byte[,] b) { //p'0 = invSR( pv(SR(p0), SR(p1)) )            
            return InvShiftRows(simpleSwap(ShiftRows(b), ShiftRows(a)));
        }
        public byte[,] RealSwapP(byte[,] a, byte[,] b) {
            return SRi(simpleSwap(SR(a), SR(b)));
        }
        public byte[,] RealSwapC(byte[,] a, byte[,] b) {// c'0 = MC◦SR(pv( SRi◦MCi(c0), SRi◦MCi(c1)))            
            return MC(SR(simpleSwap(SRi(MCi(a)), SRi(MCi(b)))));
        }
        public void swapP(ref byte[,] x, ref byte[,] y) {
            x = SR(x);
            y = SR(y);
            byte[,] a = copy(x);
            byte[,] b = copy(y);

            int z = -1;
            // Pick the first word where texts are different
            for (int i = 0; i < 4; i++) { // columns 0, 1, 2, 3
                for (int j = 0; j < 4; j++) {
                    if (x[j, i] != y[j, i]) { //Goes down each column of x and y, trying to find a difference here
                        z = i; //f.eks difference in columns 2 ,  i = 2
                        break;
                    }
                }
                if (z != -1) {
                    break; // Now we have a difference in z = i, so stop this loop and continue
                }
            }
            if (z != -1) { //if any difference
                for (int i = 0; i < 4; i++) {
                    x[i, z] = b[i, z];
                    y[i, z] = a[i, z];
                }
            }
            x = SRi(x);
            y = SRi(y);
        }
        public void swapC(ref byte[,] x, ref byte[,] y) {
            x = SRi(MCi(x));
            y = SRi(MCi(y));
            byte[,] a = copy(x);
            byte[,] b = copy(y);

            int z = -1;
            // Pick the first word where texts are different
            for (int i = 0; i < 4; i++) { // columns 0, 1, 2, 3
                for (int j = 0; j < 4; j++) {
                    if (x[j, i] != y[j, i]) { //Goes down each column of x and y, trying to find a difference here
                        z = i; //f.eks difference in columns 2 ,  i = 2
                        break;
                    }
                }
                if (z != -1) {
                    break; // Now we have a difference in z = i, so stop this loop and continue
                }
            }
            if (z != -1) { //if any difference
                for (int i = 0; i < 4; i++) {
                    x[i, z] = b[i, z];
                    y[i, z] = a[i, z];
                }
            }
            x = SR(x); x = MC(x);
            y = SR(y); y = MC(y);
        }
        public byte[,] SuperSwapP(byte[,] a, byte[,] b, byte[] v) {
            return SRi(PvC(SR(a), SR(b), v));
        }
        public byte[,] SuperSwapC(byte[,] a, byte[,] b, byte[] v) {
            return MC(SR(PvC(SRi(MCi(a)), SRi(MCi(b)), v)));
        }
        public byte[,] SuperSwapC2(byte[,] a, byte[,] b, byte[] v) {
            return MC(SR(PvD(SRi(MCi(a)), SRi(MCi(b)), v)));
        }

        // Printing--------------------------------------------------------------------------------------------------------------
        public void miniPrint(byte[] m) {
            for (int i = 0; i < 4; i++) {
                Console.Write(m[i].ToString("X").PadLeft(2, '0').ToUpper() + " ");
            }
            Console.WriteLine();
        }
        public void Print(byte[,] a) {
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    Console.Write(String.Format("{0} ", a[j, i].ToString("X").PadLeft(2, '0').ToUpper()));
                }
                Console.Write("  ");
            }
            //Console.Write("\n");
            Console.WriteLine();
        }
        public void PrintMx(byte[,] a) {
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    Console.Write(String.Format("{0} ", a[i, j].ToString("X").PadLeft(2, '0').ToUpper()));
                }
                Console.Write("\r\n");
            }
            Console.Write("\r\n\r\n");
        }
        public void printV(byte[] v) {
            Console.Write("v = [ ");
            for (int i = 0; i < 16; i++) {
                Console.Write(v[i].ToString() + ", ");
            }
            Console.Write("]\r\n");
        }
        public string ToString(byte[,] a) {
            string s = "";
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 4; j++) {
                    s += a[j, i].ToString("X").PadLeft(2, '0').ToUpper();
                }
            }
            return s;
        }
    }
}
